import numpy as np
from joblib import Parallel, delayed, cpu_count
from mne.utils import logger
from scipy import stats as st
from scipy import ndimage

from ieeg import Doubles
from ieeg.calc.reshape import make_data_same
from ieeg.calc.fast import mean_diff, permgt as permgtnd
from ieeg.process import get_mem, iterate_axes


def dist(mat: np.ndarray, axis: int = 0, mode: str = 'sem',
         where: np.ndarray = None) -> Doubles:
    """ Calculate the mean and standard deviation of a matrix.

    This function calculates the mean and standard deviation of a matrix along
    a given axis. If a mask is provided, the mean and standard deviation are
    calculated only for the elements of the matrix that are not masked.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to calculate mean and standard deviation of.
    axis : int
        Axis of matrix to calculate mean and standard deviation along.
    mode : str
        Mode of standard deviation to calculate. Can be 'sem' for standard
        error of the mean or 'std' for standard deviation.
    where : np.ndarray
        Mask of elements to include in mean and standard deviation calculation.

    Returns
    -------
    Doubles
        Tuple containing the mean and standard deviation of the matrix.

    Examples
    --------
    >>> import numpy as np
    >>> mat = np.arange(24).reshape(4,6)
    >>> dist(mat)[1] # doctest: +NORMALIZE_WHITESPACE
    array([3.87298335, 3.87298335, 3.87298335, 3.87298335, 3.87298335,
           3.87298335])
    >>> dist(mat, mode='std')[1]
    array([6.70820393, 6.70820393, 6.70820393, 6.70820393, 6.70820393,
           6.70820393])
    """

    assert mode in ('sem', 'std'), "mode must be 'sem' or 'std'"

    if where is None:
        where = np.ones(mat.shape, dtype=bool)

    where = np.logical_and(where, ~np.isnan(mat))

    mean = np.mean(mat, axis=axis, where=where)
    if mode == 'sem':
        std = st.sem(mat, axis=axis, nan_policy='omit')
    else:
        std = np.std(mat, axis=axis, where=where)

    return mean, std


def outlier_repeat(data: np.ndarray, sd: float, rounds: int = np.inf,
                   axis: int = 0) -> tuple[tuple[int, int]]:
    """ Remove outliers from data and repeat until no outliers are left.

    This function removes outliers from data and repeats until no outliers are
    left. Outliers are defined as any data point that is more than sd standard
    deviations from the mean. The function returns a tuple of tuples containing
    the index of the outlier and the round in which it was removed.

    Parameters
    ----------
    data : np.ndarray
        Data to remove outliers from.
    sd : float
        Number of standard deviations from the mean to consider an outlier.
    rounds : int
        Number of times to repeat outlier removal. If None, the function will
        repeat until no outliers are left.
    axis : int
        Axis of data to remove outliers from.

    Returns
    -------
    tuple[tuple[int, int]]
        Tuple of tuples containing the index of the outlier and the round in
        which it was removed.

    Examples
    --------
    >>> import numpy as np
    >>> data = np.array([[1, 1, 1, 1, 1], [0, 60, 0, 10, 0]]).T
    >>> tuple(outlier_repeat(data, 1))
    ((1, 1), (3, 2))
    >>> tuple(outlier_repeat(data, 1, rounds=1))
    ((1, 1),)
    >>> tuple(outlier_repeat(data, 1, rounds=0))
    ()
    """
    inds = list(range(data.shape[axis]))

    # Square the data and set zeros to small positive number
    R2 = np.square(data)
    R2[np.where(R2 == 0)] = 1e-9

    # find all axes that are not channels (example: time, trials)
    axes = tuple(i for i in range(data.ndim) if not i == axis)

    # Initialize stats loop
    sig = np.std(R2, axes)  # take standard deviation of each channel
    cutoff = (sd * np.std(sig)) + np.mean(sig)  # outlier cutoff
    i = 1

    # remove bad channels and re-calculate variance until no outliers are left
    while np.any(np.where(sig > cutoff)) and i <= rounds:

        # Pop out names to bads output using comprehension list
        for j, out in enumerate(np.where(sig > cutoff)[0]):
            yield inds.pop(out - j), i

        # re-calculate per channel variance
        R2 = R2[..., np.where(sig < cutoff)[0], :]
        sig = np.std(R2, axes)
        cutoff = (sd * np.std(sig)) + np.mean(sig)
        i += 1


def find_outliers(data: np.ndarray, outliers: float) -> np.ndarray[bool]:
    """ Find outliers in data matrix.

    This function finds outliers in a data matrix. Outliers are defined as any
    trial with a maximum value greater than the mean plus outliers times the
    standard deviation. The function returns a boolean array with True for
    trials that are not outliers and False for trials that are outliers.

    Parameters
    ----------
    data : np.ndarray
        Data to find outliers in.
    outliers : float
        Number of standard deviations from the mean to consider an outlier.

    Returns
    -------
    np.ndarray[bool]
        Boolean array with True for trials that are not outliers and False
        for trials that are outliers.

    Examples
    --------
    >>> import numpy as np
    >>> data = np.array([[1, 1, 1, 1, 1], [0, 60, 0, 10, 0]]).T
    >>> find_outliers(data, 1)
    array([ True, False,  True,  True,  True])
    >>> find_outliers(data, 3)
    array([ True,  True,  True,  True,  True])
    >>> find_outliers(data, 0.1)
    array([ True, False,  True, False,  True])
    """
    dat = np.abs(data)  # (trials X channels X (frequency) X time)
    max = np.max(dat, axis=-1)  # (trials X channels X (frequency))
    std = np.std(dat, axis=(-1, 0))  # (channels X (frequency))
    mean = np.mean(dat, axis=(-1, 0))  # (channels X (frequency))
    keep = max < ((outliers * std) + mean)  # (trials X channels X (frequency))
    return keep


def avg_no_outlier(data: np.ndarray, outliers: float = None,
                   keep: np.ndarray[bool] = None) -> np.ndarray:
    """Calculate the average of data without trial outliers.

    This function calculates the average of data without trial outliers.
    Outliers are defined as any trial with a maximum value greater than the
    mean plus outliers times the standard deviation.
    The function returns the average of data without outliers.

    Parameters
    ----------
    data : np.ndarray
        Data to calculate average of.
    outliers : float
        Number of standard deviations from the mean to consider an outlier.
    keep : np.ndarray[bool]
        Boolean array with True for trials that are not outliers and False
        for trials that are outliers.

    Returns
    -------
    np.ndarray
        Average of data without outliers.

    Examples
    --------
    >>> import mne
    >>> mne.set_log_file(None)
    >>> data = np.array([[[1, 1, 1, 1, 1], [0, 60, 0, 10, 0]]]).T
    >>> avg_no_outlier(data, 1)
    Removed Trial 0 in Channel 0
    Removed Trial 1 in Channel 0
    Removed Trial 1 in Channel 1
    Removed Trial 2 in Channel 0
    Removed Trial 3 in Channel 0
    Removed Trial 4 in Channel 0
    array([[nan],
           [2.5]])
    >>> avg_no_outlier(data, 3)
    Removed Trial 0 in Channel 0
    Removed Trial 1 in Channel 0
    Removed Trial 2 in Channel 0
    Removed Trial 3 in Channel 0
    Removed Trial 4 in Channel 0
    array([[nan],
           [14.]])
    """
    if data.ndim not in (3, 4):
        raise ValueError("Data must be 3D or 4D")

    if keep is None and outliers is not None:
        keep = find_outliers(data, outliers)
    elif keep is not None:
        if keep.ndim == 2 and data.ndim == 4:
            keep = keep[..., np.newaxis]
        elif keep.ndim == data.ndim:
            raise ValueError(f"Keep has too many dimensions ({keep.ndim})")
    else:
        raise ValueError("Either keep or outliers must be given")

    if np.squeeze(keep).ndim == 2:  # dat.ndim == 3
        disp = [f"Removed Trial {i} in Channel {j}" for i, j in np.ndindex(
            data.shape[0:2]) if not keep[i, j]]
    else:  # dat.ndim == 4:
        disp = [f"Removed Trial {i} in Channel {j} in Frequency {k}" for i, j,
                k in np.ndindex(data.shape[0:3]) if not keep[i, j, k]]
    for msg in disp:
        logger.info(msg)
    return np.mean(data, axis=0, where=keep[..., np.newaxis])

# def perm_test(group1: np.ndarray, group2: np.ndarray, n_perm: int = 1000,
#               axis: int=-1):
#     """Calculate the permutation test.
#
#     This function calculates the permutation test between two groups of
#     observations. The permutation test is a non-parametric test that compares
#     the mean difference between two groups of observations to the mean
#     difference between two groups of observations that have been randomly
#     permuted. The function returns the p-value of the observed difference.
#
#     Parameters
#     ----------
#     group1 : array, shape (..., time)
#         The first group of observations.
#     group2 : array, shape (..., time)
#         The second group of observations.
#     n_perm : int, optional
#         The number of permutations to perform.
#     axis : int, optional
#         The axis along which to calculate the statistic function.
#
#     Returns
#     -------
#     p : float
#         The p-value of the observed difference.
#
#     Examples
#     --------
#     >>> import numpy as np
#     >>> seed = 43; rng = np.random.default_rng(seed)
#     >>> sig1 = np.array([[0,1,1,2,2,2.5,3,3,3,2.5,2,2,1,1,0]
#     ... for _ in range(50)])
#     >>> sig2 = rng.random((100, 15)) * 2
#     >>> perm_test(sig1, sig2, n_perm=1000000, axis=0) # doctest: +ELLIPSIS
#     0.0
#     """
#     in1 = np.moveaxis(group1, axis, -1).astype('double')
#     in2 = np.moveaxis(group2, axis, -1).astype('double')
#
#     return _perm_test(in1, in2, n_perm)


def window_averaged_shuffle(sig1: np.ndarray, sig2: np.ndarray,
                            n_perm: int = 100000, tails: int = 1,
                            obs_axis: int = 0, window_axis: int = -1,
                            stat_func: callable = mean_diff, seed: int = None,
                            ) -> np.ndarray[bool]:
    """Calculate the window averaged shuffle distribution.

    Essentially a wrapper for:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.permutatio
    n_test.html#scipy.stats.permutation_test

    Parameters
    ----------
    sig1 : array, shape (trials, ..., time)
        The first group of observations.
    sig2 : array, shape (trials, ..., time)
        The second group of observations.
    n_perm : int, optional
        The number of permutations to perform. Default is 1000.
    tails : int, optional
        The number of tails to use for the p-value. Default is 1.
    obs_axis : int, optional
        The axis along which to calculate the statistic function. Default is 0.
    window_axis : int, optional
        The axis along which to calculate the window average. Default is -1.
    stat_func : callable, optional
        The statistic function to use. Default is mean_diff.
    seed : int, optional
        The random seed to use for the permutation test. Default is None.

    Returns
    -------
    shuffle_dist : np.ndarray
        The shuffle distribution.

    Examples
    --------
    >>> import numpy as np
    >>> seed = 43; rng = np.random.default_rng(seed)
    >>> sig1 = np.array([[0,1,1,2,2,2.5,3,3,3,2.5,2,2,1,1,0]
    ... for _ in range(50)])
    >>> sig2 = rng.random((100, 15)) * 3.2
    >>> window_averaged_shuffle(sig1, sig2, n_perm=1000000, seed=seed
    ... ) # doctest: +ELLIPSIS
    0.99...
    """

    # average the windows
    samples = [np.atleast_1d(np.mean(s, axis=window_axis))
               for s in [sig1, sig2]]

    # calc obs axis
    obs_axis = obs_axis + sig1.ndim if obs_axis < 0 else obs_axis
    window_axis = window_axis + sig1.ndim if window_axis < 0 else window_axis
    obs_axis = obs_axis - 1 if window_axis < obs_axis else obs_axis

    if tails == -1:
        alt = 'greater'
    elif tails == 1:
        alt = 'less'
    else:
        alt = 'two-sided'
    out_mem = (sig1.size + sig2.size) * 8
    batch_size = get_mem() // out_mem

    # Create shuffle distribution
    res = st.permutation_test(samples, stat_func,
                              n_resamples=n_perm,
                              alternative=alt,
                              batch=batch_size,
                              axis=obs_axis,
                              random_state=seed)

    return res.pvalue


def time_perm_cluster(sig1: np.ndarray, sig2: np.ndarray, p_thresh: float,
                      p_cluster: float = None, n_perm: int = 1000,
                      tails: int = 1, axis: int = 0,
                      stat_func: callable = mean_diff,
                      ignore_adjacency: tuple[int] | int = None,
                      n_jobs: int = -1, seed: int = None
                      ) -> (np.ndarray[bool], np.ndarray[float]):
    """Calculate significant clusters using permutation testing and cluster
    correction.

    Takes two time series signals, finding clusters of activation defined as
    significant contiguous time points with a p value less than the p_thresh
    (greater if tails is -1, and both if tails is 2). The p value is the
    proportion of times that the difference in the statistic value for signal 1
    with respect to signal 2 is greater than the statistic for a random
    sampling of signal 1 and 2 with respect to a random sampling of signal 1
    and 2. The clusters are then corrected by only keeping clusters that are in
    the (1 - p_cluster)'th percentile of cluster lengths for signal 2.

    Parameters
    ----------
    sig1 : array, shape (trials, ..., time)
        Active signal. The first dimension is assumed to be the trials
    sig2 : array, shape (trials, ..., time)
        Passive signal. The first dimension is assumed to be the trials
    p_thresh : float
        The p-value threshold to use for determining significant time points.
    p_cluster : float, optional
        The p-value threshold to use for determining significant clusters.
    n_perm : int, optional
        The number of permutations to perform.
    tails : int, optional
        The number of tails to use. 1 for one-tailed, 2 for two-tailed.
    axis : int, optional
        The axis to perform the permutation test across. Also known as the
        observations axis
    stat_func : callable, optional
        The statistical function to use to compare populations. Requires an
        axis keyword input to denote observations (trials, for example).
        Default function is `mean_diff`, but may be substituted with other test
        functions found here:
        https://scipy.github.io/devdocs/reference/stats.html#independent
        -sample-tests
    ignore_adjacency : int or tuple of ints, optional
        The axis or axes to ignore when finding clusters. For example, if
        sig1.shape = (trials, channels, time), and you want to find clusters
        across time, but not channels, you would set ignore_adjacency = 1.
    n_jobs : int, optional
        The number of jobs to run in parallel. -1 for all processors. Default
        is -1.
    seed : int, optional
        The random seed to use for the permutation test. Default is None.

    Returns
    -------
    clusters : array, shape (..., time)
        The binary array of significant clusters.
    p_obs : array, shape (..., time)
        The p-value of the observed difference.

    References
    ----------
    1. https://www.sciencedirect.com/science/article/pii/S0165027007001707

    Examples
    --------
    >>> import numpy as np
    >>> seed = 43; rng = np.random.default_rng(seed)
    >>> sig1 = np.array([[0,1,1,2,2,2.5,3,3,3,2.5,2,2,1,1,0]
    ... for _ in range(50)]) - rng.random((50, 15)) * 2.6
    >>> sig2 = np.array([[0] * 15 for _ in range(100)]) + rng.random(
    ... (100, 15))
    >>> time_perm_cluster(sig1, sig2, 0.05, n_perm=100000, seed=seed)[0]
    array([False, False, False,  True,  True,  True,  True,  True,  True,
            True,  True, False, False, False, False])
    >>> time_perm_cluster(sig1, sig2, 0.01, n_perm=100000, seed=seed)[0]
    array([False, False, False, False,  True,  True,  True,  True,  True,
            True, False, False, False, False, False])
    """
    # check inputs
    if tails == 1:
        alt = 'greater'
    elif tails == -1:
        alt = 'less'
    elif tails == 2:
        alt = 'two-sided'
    else:
        raise ValueError('tails must be 1, 2, or -1')
    if p_cluster is None:
        p_cluster = p_thresh
    if p_cluster > 1 or p_cluster < 0 or p_thresh > 1 or p_thresh < 0:
        raise ValueError('p_thresh and p_cluster must be between 0 and 1')
    if isinstance(ignore_adjacency, int):
        ignore_adjacency = (ignore_adjacency,)
    if isinstance(ignore_adjacency, tuple):
        assert axis not in ignore_adjacency, ValueError(
            'observations axis is eliminated before clustering and so cannot '
            'be in ignore_adjacency')
        nprocs = sum(sig1.shape[i] for i in ignore_adjacency)
        n_jobs += cpu_count() + 1 if n_jobs < 0 else 0
    else:
        nprocs = 1

    # set process parameters
    sig2 = make_data_same(sig2, sig1.shape, axis)
    sample_size = sig1.nbytes + sig2.nbytes
    batch_size = get_mem() // sample_size
    small_enough = batch_size > n_perm ** 2
    kwargs = dict(n_resamples=n_perm, alternative=alt, batch=batch_size,
                  axis=axis, vectorized=True)

    if isinstance(stat_func([1], [1]), tuple):
        logger.warning('stat_func returns a tuple. Taking the first element')
        func = stat_func

        def stat_func(*args, **kwargs):
            return func(*args, **kwargs)[0]

    # Create binary clusters using the p value threshold
    def _proc(sig1: np.ndarray, sig2: np.ndarray
              ) -> tuple[np.ndarray[int], np.ndarray[float]]:
        res = st.permutation_test([sig1, sig2], stat_func, **kwargs)
        p_act = res.pvalue
        diff = res.null_distribution

        # Calculate the p value of the permutation distribution
        if small_enough:
            p_perm = np.mean(tail_compare(diff[:, None], diff[None], tails),
                             axis=0)
        elif tails == 1:
            p_perm = permgtnd(diff, axis=0)
        else:
            p_perm = proportion(diff, tail=tails, axis=0)

        # Create binary clusters using the p value threshold
        b_act = tail_compare(1. - p_act, 1. - p_thresh, tails)
        b_perm = tail_compare(p_perm, 1. - p_thresh, tails)

        return time_cluster(b_act, b_perm, 1 - p_cluster, tails), p_act

    # axes where adjacency is ignored can be computed independently in
    # parallel
    if nprocs > 1:
        out_shape = sig1.shape[:axis] + sig1.shape[axis + 1:]
        out1 = np.zeros(out_shape, dtype=int)
        out2 = np.zeros(out_shape, dtype=float)
        ins = ((sig1[i], sig2[i]) for i in iterate_axes(
            sig1, ignore_adjacency))
        axis -= sum(1 for i in ignore_adjacency if i < axis)
        proc = Parallel(n_jobs, return_as='generator', verbose=40,)(
            delayed(_proc)(*i) for i in ins)
        for i, iout in enumerate(proc):
            out1[i], out2[i] = iout
        return out1, out2
    else:
        return _proc(sig1, sig2)


def proportion(val: np.ndarray[float, ...] | float,
               comp: np.ndarray[float, ...] = None, tail: int = 1,
               axis: int = None) -> np.ndarray[float, ...] | float:
    """takes a value and a comparison and returns the proportion of the
    comparison that is greater than the value

    Parameters
    ----------
    val : array, shape (x, ...) or float,
        The difference between two groups.
    comp : array, shape (y, ...) optional
        The difference between two groups.
    tail : int, optional
        The number of tails to use. 1 for one-tailed, 2 for two-tailed.
    axis : int, optional
        The axis to perform the permutation test across. Also known as the
        observations axis

    Returns
    -------
    proportion : array, shape (..., time)
        The proportion of the comparison that is greater than the value.

    Examples
    ________
    >>> import numpy as np
    >>> rand = np.random.default_rng(seed=42)
    >>> diff1 = rand.random(5)
    >>> diff1
    array([0.77395605, 0.43887844, 0.85859792, 0.69736803, 0.09417735])
    >>> proportion(diff1)
    array([0.75, 0.25, 1.  , 0.5 , 0.  ])
    >>> np.sum(diff1 > diff1[:, None], axis=0) / (diff1.shape[0] - 1)
    array([0.75, 0.25, 1.  , 0.5 , 0.  ])
    >>> diff2 = rand.random((2, 4))
    >>> diff2
    array([[0.97562235, 0.7611397 , 0.78606431, 0.12811363],
           [0.45038594, 0.37079802, 0.92676499, 0.64386512]])
    >>> proportion(diff2, axis=1)
    array([[1.        , 0.33333333, 0.66666667, 0.        ],
           [0.33333333, 0.        , 1.        , 0.66666667]])
    >>> proportion(diff2, axis=0)
    array([[1., 1., 0., 0.],
           [0., 0., 1., 1.]])
    >>> val = 0.5
    >>> compare = np.array([0.2, 0.4, 0.5, 0.7, 0.9])
    >>> proportion(compare)
    array([0.  , 0.25, 0.5 , 0.75, 1.  ])
    >>> val = np.full(5, 0.5)
    >>> compare = np.array([[0.2, 0.4, 0.4, 0.7, 0.9],
    ...                     [0.1, 0.3, 0.6, 0.5, 0.9]])
    >>> proportion(val, compare, axis=0) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    NotImplementedError
    """

    match tail:
        case 1:
            pass
        case 2:
            val = np.abs(val)
            if comp is not None:
                comp = np.abs(comp)
        case -1:
            val *= -1
            if comp is not None:
                comp *= -1
        case _:
            raise ValueError('tail must be 1, 2, or -1')

    if axis is None and comp is None:
        return _perm_gt_1d(val)
    elif comp is None:
        return _perm_gt_1d(val, axis=axis)
    else:
        raise NotImplementedError()


# @guvectorize(['void(f8[::1], f8[::1])'], '(n)->(n)', nopython=True)
def _perm_gt_1d(diff, axis=0):
    m = diff.shape[axis] - 1
    sorted_indices = diff.argsort(axis=axis)  # Get sorted indices
    proportions = np.arange(diff.shape[axis]) / m  # Create proportions array
    # Rearrange to match original order
    return proportions[sorted_indices.argsort(axis=axis)]


def time_cluster(act: np.ndarray, perm: np.ndarray, p_val: float = None,
                 tails: int = 1) -> np.ndarray:
    """Cluster correction for time series data.

    1. Creates an index of all the binary clusters in the active and permuted
        passive data.
    2. For each permutation in the passive data, determine the maximum cluster
        size.
    3. For each cluster in the active data, determine the proportion of
        permutations that have a cluster of the same size or larger.

    Parameters
    ----------

    act : array, shape (time)
        The active data.
    perm : array, shape (n_perm, time)
        The permutation passive data.
    p_val : float, optional
        The p-value threshold to use for determining significant clusters.
    tails : int, optional
        The number of tails to use. 1 for one-tailed, 2 for two-tailed.

    Returns
    -------
    clusters : array, shape (..., time)
        The clusters.

    Examples
    --------
    >>> import numpy as np
    >>> perm = np.array([[0, 1, 1, 1, 0, 1, 0, 0],
    ...                  [0, 1, 0, 1, 1, 1, 0, 0],
    ...                  [0, 1, 1, 0, 1, 1, 0, 0],
    ...                  [0, 0, 1, 1, 1, 0, 0, 0]])
    >>> time_cluster(np.array([0, 1, 1, 1, 1, 1, 0, 0]), perm)
    array([0.  , 0.75, 0.75, 0.75, 0.75, 0.75, 0.  , 0.  ])
    >>> time_cluster(np.array([0, 0, 1, 1, 1, 0, 0, 0]), perm)
    array([0.  , 0.  , 0.25, 0.25, 0.25, 0.  , 0.  , 0.  ])
    """

    # Create an index of all the binary clusters in the active and permuted
    # passive data
    footprint = ndimage.generate_binary_structure(act.ndim, 1)
    perm_footprint = np.stack((np.zeros_like(footprint), footprint,
                               np.zeros_like(footprint)))
    act_clusters = ndimage.label(act, footprint)[0]
    perm_clusters = ndimage.label(perm, perm_footprint)[0]
    unique, idx, counts = np.unique(perm_clusters,
                                    return_index=True, return_counts=True)
    idxs = np.unravel_index(idx, perm_clusters.shape)
    max_cluster_len = np.zeros(perm_clusters.shape[0])
    for i in np.unique(idxs[0]):
        max_cluster_len[i] = np.max(counts[idxs[0] == i])

    # For each cluster in the active data, determine the proportion of
    # permutations that have a cluster of the same size or larger
    cluster_p_values = np.zeros(act_clusters.shape)
    for i in range(1, act_clusters.max() + 1):
        # Get the cluster
        act_cluster = act_clusters == i
        # Get the cluster size
        act_cluster_size = np.sum(act_cluster)
        # Determine the proportion of permutations that have a cluster of the
        # same size or larger
        cluster_p_values[act_cluster] = np.mean(act_cluster_size >
                                                max_cluster_len, axis=0)

    # If p_val is not None, return the boolean array indicating whether the
    # cluster is significant
    if p_val is not None:
        return tail_compare(cluster_p_values, p_val, tails)
    else:
        return cluster_p_values


def tail_compare(diff: np.ndarray | float | int,
                 obs_diff: np.ndarray | float | int, tails: int = 1
                 ) -> np.ndarray | bool:
    """Compare the difference between two groups to the observed difference.

    This function applies the appropriate comparison based on the number of
    tails. The shapes of the two arrays must be broadcastable.

    Parameters
    ----------
    diff : array
        The difference between the two groups.
    obs_diff : array
        The observed difference between the two groups.
    tails : int, optional
        The number of tails to use. 1 for one-tailed, 2 for two-tailed.

    Returns
    -------
    larger : array, shape (..., time)
        The boolean array indicating whether the difference between the two
        groups is larger than the observed difference.

    Examples
    --------
    >>> import numpy as np
    >>> rand = np.random.default_rng(seed=42)
    >>> diff1 = rand.random(5) - 0.5
    >>> diff1
    array([ 0.27395605, -0.06112156,  0.35859792,  0.19736803, -0.40582265])
    >>> obs_diff1 = 0.25
    >>> tail_compare(diff1, obs_diff1)
    array([ True, False,  True, False, False])
    >>> tail_compare(diff1, obs_diff1, tails=2)
    array([ True, False,  True, False,  True])
    >>> tail_compare(diff1, obs_diff1, tails=-1)
    array([False,  True, False,  True,  True])
    >>> tail_compare(diff1, np.array([1, 2])
    ... ) # doctest: +ELLIPSIS +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ValueError: shape mismatch: objects cannot be broadcast to a single ...
    """

    # check if arrays are broadcastable
    try:
        np.broadcast(diff, obs_diff)
    except ValueError as e:
        raise e

    # Account for one or two tailed test
    match tails:
        case 1:
            temp = np.greater(diff, obs_diff)
        case 2:
            temp = np.greater(np.abs(diff), np.abs(obs_diff))
        case -1:
            temp = np.less(diff, obs_diff)
        case _:
            raise ValueError('tails must be 1, 2, or -1')

    return temp


# def shuffle_test(sig1: np.ndarray, sig2: np.ndarray, n_perm: int = 1000,
#                  axis: int = 0, func: callable = mean_diff, seed: int = None,
#                  n_jobs: int = -3) -> np.ndarray:
#     """Time permutation shuffle test between two set of observations.
#
#     The test is performed by shuffling the trials and calculating the test
#     statistic (the group difference by default). The shuffling together of
#     groups should create a null distribution of values to compare against for
#     statistical significance. The number of trials in each group does not nee
#     to be the same, but all other dimensions must be.
#
#     Parameters
#     ----------
#     sig1 : array, shape (trials1, ...)
#         Active signal. The first dimension is assumed to be the trials
#     sig2 : array, shape (trials2, ...)
#         Passive signal. The first dimension is assumed to be the trials
#     n_perm : int, optional
#         The number of permutations to perform.
#     axis : int, optional
#         The axis to perform the permutation test across (trials).
#     func :
#         The statistical function to use to compare populations. Requires an
#         axis keyword input to denote observations (trials, for example).
#     seed : int, optional
#         The seed for the random number generator.
#     n_jobs : int, optional
#         The number of jobs to run in parallel. Only used if the permutation
#         test will exceed memory. Default is -3.
#
#     Returns
#     -------
#     p : np.ndarray, shape (...)
#         The p-values
#
#     Examples
#     --------
#     >>> import numpy as np
#     >>> seed = 42; rng = np.random.default_rng(seed)
#     >>> sig1 = np.mean(np.array([[0,1,1,2,2,2.5,3,3,3,2.5,2,2,1,1,0]]),
#     axis=1)
#     >>> sig2 = np.mean(rng.random((100, 15)) * 2.4, axis=1)
#     >>> round(np.mean(sig1 - sig2), 3)
#     0.534
#     >>> np.mean(np.mean(sig1 - sig2) > shuffle_test(sig1, sig2,
#     n_perm=1000000,
#     ... seed=seed))
#     0.993
#     """
#
#     rng = np.random.default_rng(seed)
#     axis = axis + sig1.ndim if axis < 0 else axis
#
#     # Concatenate the two signals for trial shuffling
#     all_trial = np.concatenate((sig1, sig2), axis=axis)
#     shape = sig1.shape
#
#     # Generate all permutations at once
#     idx = np.tile(np.arange(all_trial.shape[axis]), (n_perm, 1))
#     rng.permuted(idx, axis=1, out=idx)
#     idx1 = idx[:, :shape[axis]]
#     idx2 = idx[:, shape[axis]:]
#
#     # Check if the permutation will exceed memory
#     out_mem = all_trial.size * n_perm * 8
#     if out_mem > get_mem() * psutil.cpu_count():
#         logger.warning(f"Permutation test will exceed memory ({out_mem}
#         bytes)"
#                        f", using a generator instead. This may take longer.")
#
#         def _shuffle_test(idx_1, idx_2):
#             fake_sig1 = np.take(all_trial, idx_1, axis=axis)
#             fake_sig2 = np.take(all_trial, idx_2, axis=axis)
#             return func(fake_sig1, fake_sig2, axis=axis)
#
#         diff = np.zeros((n_perm, *shape[:axis], *shape[axis + 1:]))
#         proc = Parallel(n_jobs=n_jobs, verbose=40)(delayed(_shuffle_test)(
#             idx1[i], idx2[i]) for i in range(n_perm))
#         for i, out in enumerate(proc):
#             diff[i] = out
#     else:
#
#         # Split the permuted trials into fake_sig1 and fake_sig2
#         fake_sig1 = np.take(all_trial, idx1, axis=axis)
#         fake_sig2 = np.take(all_trial, idx2, axis=axis)
#
#         # Calculate the average difference between the two groups averaged
#         # across trials
#         diff = func(fake_sig1, fake_sig2, axis=axis + 1)
#
#     return diff


def sum_squared(x: np.ndarray) -> np.ndarray | float:
    """Compute norm of an array.

    Parameters
    ----------
    x : array
        Data whose norm must be found.
    Returns
    -------
    value : float
        Sum of squares of the input array x.
    """
    x_flat = x.ravel()
    return np.dot(x_flat, x_flat)


def sine_f_test(window_fun: np.ndarray, x_p: np.ndarray
                ) -> (np.ndarray, np.ndarray):
    """Computes the F-statistic for sine wave in locally-white noise.

    This function computes the F-statistic for a sine wave in locally-white
    noise. The sine wave is assumed to be of the form:

    :math:`x(t) = A \\sin(2 \\pi f t + \\phi)`

    where :math:`A` is the amplitude of the sine wave, :math:`f` is the
    frequency of the sine wave, and :math:`\\phi` is the phase of the sine
    wave. The F-statistic is computed by taking the ratio of the variance of
    the sine wave to the variance of the noise. The variance of the sine wave
    is computed by taking the sum of the squares of the sine wave across tapers
    and then dividing by the number of tapers. The variance of the noise is
    computed by taking the sum of the squares of the residuals across tapers
    and then dividing by the number of tapers minus one. The F-statistic is
    then computed by taking the ratio of the variance of the sine wave to the
    variance of the noise.

    Parameters
    ----------
    window_fun : array
        The tapers used to calculate the multitaper spectrum.
    x_p : array
        The tapered time series.

    Returns
    -------
    f_stat : array
        The F-statistic for each frequency.
    A : array
        The amplitude of the sine wave at each frequency.

    Examples:
    ---------
    >>> import numpy as np
    >>> window_fun = np.array([[0.5, 0.5], [0.5, -0.5]])
    >>> x_p = np.array([[[1, 1, 1, 1, 1], [0, 60, 0, 10, 0]],
    ...                 [[1, 1, 1, 1, 1], [0, 0, 0, 0, 0]]]).T
    >>> sine_f_test(window_fun, x_p)
    (array([[0.        , 0.        ],
           [0.00027778, 0.        ],
           [0.        , 0.        ],
           [0.01      , 0.        ],
           [0.        , 0.        ]]), array([[1., 1.],
           [1., 1.],
           [1., 1.],
           [1., 1.],
           [1., 1.]]))

    """
    # drop the even tapers
    n_tapers = len(window_fun)
    tapers_odd = np.arange(0, n_tapers, 2)
    tapers_even = np.arange(1, n_tapers, 2)
    tapers_use = window_fun[tapers_odd]

    # sum tapers for (used) odd prolates across time (n_tapers, 1)
    H0 = np.sum(tapers_use, axis=1)

    # sum of squares across tapers (1, )
    H0_sq = sum_squared(H0)

    # sum of the product of x_p and H0 across tapers (1, n_freqs)
    exp_H0 = np.reshape(H0, (1, -1, 1))
    x_p_H0 = np.sum(x_p[:, tapers_odd, :] * exp_H0, axis=1)

    # resulting calculated amplitudes for all freqs
    A = x_p_H0 / H0_sq

    # figure out which freqs to remove using F stat

    # estimated coefficient
    x_hat = A * np.reshape(H0, (-1, 1))

    # numerator for F-statistic
    num = (n_tapers - 1) * (A * A.conj()).real * H0_sq
    # denominator for F-statistic
    den = (np.sum(np.abs(x_p[:, tapers_odd, :] - x_hat) ** 2, 1) +
           np.sum(np.abs(x_p[:, tapers_even, :]) ** 2, 1))
    den = np.where(den == 0, np.inf, den)
    f_stat = num / den

    return f_stat, A


if __name__ == '__main__':
    import numpy as np
    from timeit import timeit

    rng = np.random.default_rng(seed=42)
    sig1 = np.array([[0, 1, 2, 3, 3] for _ in range(50)]) - rng.random(
        (50, 5)) * 5
    sig2 = np.array([[0] * 5 for _ in range(100)]) + rng.random((100, 5))
    diff = shuffle_test(sig1, sig2, 3000, 0)
    act = mean_diff(sig1, sig2, axis=0)

    # Calculate the p value of the permutation distribution and compare
    # execution times

    # p_perm1 = _perm_gt(diff, diff)
    p_perm2 = np.sum(diff[None] > diff[:, None], axis=0) / (diff.shape[0] - 1)
    # p_perm3 = (_perm_gt(diff, diff[:, None], axis=0) * diff.shape[0] /
    #            (diff.shape[0] - 1))
    p_perm4 = proportion(diff, axis=0)

    # Time the functions
    runs = 20
    # time1 = timeit('_perm_gt(diff, diff)', globals=globals(), number=runs)
    time2 = timeit('np.sum(diff > diff[:, np.newaxis], axis=0) / '
                   '(diff.shape[0] - 1)', globals=globals(), number=runs)
    # time3 = timeit('_perm_gt(diff[:, None], diff, axis=0) * diff.shape[0]'
    #                '/ (diff.shape[0] - 1)', globals=globals(), number=runs)
    time4 = timeit('proportion(diff, axis=0)', globals=globals(), number=runs)

    # print(f'Time for _perm_gt_2: {time1 / runs:.6f} seconds per run')
    print(f'Time for sum method: {time2 / runs:.6f} seconds per run')
    # print(f'Time for _perm_gt: {time3 / runs:.6f} seconds per run')
    print(f'Time for perm_gt: {time4 / runs:.6f} seconds per run')



import csv
import os.path as op
from collections import OrderedDict
from collections.abc import Iterable, Sequence
from functools import singledispatch

import mne
import nibabel as nib
import numpy as np
from mne.viz import Brain

from ieeg import PathLike, Signal
from ieeg.io import get_elec_volume_labels
from ieeg.viz import _qt_backend, parula

_qt_backend()

import matplotlib  # noqa: E402
import matplotlib.patheffects as path_effects  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402


def plot_overlay(image: nib.Nifti1Image, compare: nib.Nifti1Image,
                 title: str, thresh: float = None):
    """Plots an overlay of two images

    Parameters
    ----------
    image : nib.Nifti1Image
        The image to plot
    compare : nib.Nifti1Image
        The image to overlay
    title : str
        The title of the plot
    thresh : float, optional
        The threshold to apply to the overlay, by default None
    """
    image = nib.orientations.apply_orientation(
        np.asarray(image.dataobj), nib.orientations.axcodes2ornt(
            nib.orientations.aff2axcodes(image.affine))).astype(np.float32)
    compare = nib.orientations.apply_orientation(
        np.asarray(compare.dataobj), nib.orientations.axcodes2ornt(
            nib.orientations.aff2axcodes(compare.affine))).astype(np.float32)
    if thresh is not None:
        compare[compare < np.quantile(compare, thresh)] = np.nan
    fig, axes = plt.subplots(1, 3, figsize=(12, 4))
    fig.suptitle(title)
    for i, ax in enumerate(axes):
        ax.imshow(np.take(image, [image.shape[i] // 2], axis=i).squeeze().T,
                  cmap='gray')
        ax.imshow(np.take(compare, [compare.shape[i] // 2],
                          axis=i).squeeze().T, cmap='gist_heat', alpha=0.5)
        ax.invert_yaxis()
        ax.axis('off')
    fig.tight_layout()


def allign_CT(t1_path: PathLike, ct_path: PathLike, reg_affine=None
              ) -> nib.spatialimages.SpatialImage:
    """Alligns a CT scan to a T1 scan

    Parameters
    ----------
    t1_path : PathLike
        The path to the T1 scan
    ct_path : PathLike
        The path to the CT scan
    reg_affine : np.ndarray, optional
        The affine to use for registration, by default None

    Returns
    -------
    nib.spatialimages.SpatialImage
        The alligned CT scan
    """
    T1 = nib.load(t1_path)
    CT_orig = nib.load(ct_path)
    sdr_morph = None
    if reg_affine is None:
        reg_affine, sdr_morph = mne.transforms.compute_volume_registration(
            CT_orig, T1, pipeline='all')
    CT_aligned = mne.transforms.apply_volume_registration(CT_orig, T1,
                                                          reg_affine,
                                                          sdr_morph)
    return CT_aligned


def show_brain(my_raw: Signal, trans: mne.transforms.Transform,
               sub_id: PathLike, subjects_dir: PathLike = None,
               overwrite: bool = False):
    """Shows the brain with the electrodes projected onto it

    Parameters
    ----------
    my_raw : Signal
        The data to plot
    trans : mne.transforms.Transform
        The transformation to use
    sub_id : PathLike
        The subject id
    subjects_dir : PathLike, optional
        The subjects directory, by default LAB_root / 'ECoG_Recon_Full'
    overwrite : bool, optional
        Whether to overwrite the watershed bem, by default False
    """
    subjects_dir = get_sub_dir(subjects_dir)
    try:
        mne.bem.make_watershed_bem(sub_id, subjects_dir, overwrite=overwrite)
    except RuntimeError:
        pass
    my_raw.info = mne.preprocessing.ieeg.project_sensors_onto_brain(
        my_raw.info, trans, sub_id, subjects_dir=subjects_dir)
    brain_kwargs = dict(cortex='low_contrast', alpha=0.2, background='white')
    brain = mne.viz.Brain(sub_id, surf='white', subjects_dir=subjects_dir,
                          title='Projection', **brain_kwargs)
    brain.add_sensors(my_raw.info, trans=trans)
    view_kwargs = dict(azimuth=60, elevation=100, distance=350,
                       focalpoint=(0, 0, -15))
    brain.show_view(**view_kwargs)


def imshow_mri(data, img: nib.spatialimages.SpatialImage,
               vox: tuple[int, int, int], xyz: dict, suptitle: str = ""):
    """Show an MRI slice with a voxel annotated.

    Parameters
    ----------
    data : np.ndarray
        The data to plot
    img : nib.spatialimages.SpatialImage
        The image to plot
    vox : tuple[int, int, int]
        The voxel to annotate
    xyz : dict
        The xyz coordinates of the voxel
    suptitle : str, optional
        The title of the plot, by default ""

    Returns
    -------
    fig, ax : matplotlib.pyplot.Figure, matplotlib.pyplot.Axes
        The figure and axes of the plot
    """
    i, j, k = vox
    fig, ax = plt.subplots(1, figsize=(6, 6))
    codes = nib.orientations.aff2axcodes(img.affine)
    # Figure out the title based on the code of this axis
    ori_slice = dict(
        P="Coronal", A="Coronal", I="Axial", S="Axial", L="Sagittal",
        R="Sagittal"
    )
    ori_names = dict(
        P="posterior", A="anterior", I="inferior", S="superior", L="left",
        R="right"
    )
    title = ori_slice[codes[0]]
    ax.imshow(data[i], vmin=10, vmax=120, cmap="gray", origin="lower")
    ax.axvline(k, color="y")
    ax.axhline(j, color="y")
    for kind, coords in xyz.items():
        annotation = "{}: {}, {}, {} mm".format(kind,
                                                *np.round(coords).astype(int))
        text = ax.text(k, j, annotation, va="baseline", ha="right",
                       color=(1, 1, 0.7))
        text.set_path_effects(
            [
                path_effects.Stroke(linewidth=2, foreground="black"),
                path_effects.Normal(),
            ]
        )
    # reorient view so that RAS is always rightward and upward
    x_order = -1 if codes[2] in "LIP" else 1
    y_order = -1 if codes[1] in "LIP" else 1
    ax.set(
        xlim=[0, data.shape[2] - 1][::x_order],
        ylim=[0, data.shape[1] - 1][::y_order],
        xlabel=f"k ({ori_names[codes[2]]}+)",
        ylabel=f"j ({ori_names[codes[1]]}+)",
        title=f"{title} view: i={i} ({ori_names[codes[0]]}+)",
    )
    fig.suptitle(suptitle)
    fig.subplots_adjust(0.1, 0.1, 0.95, 0.85)
    return fig


def get_sub_dir(subj_dir: PathLike = None):
    """Gets the subjects directory

    Parameters
    ----------
    subj_dir : PathLike, optional
        The subjects directory, by default LAB_root / 'ECoG_Recon_Full'

    Returns
    -------
    PathLike
        The subjects directory
    """
    if subj_dir is None:
        from os import path
        HOME = path.expanduser("~")
        subj_dir = op.join(HOME, "Box", "ECoG_Recon")
    return subj_dir


def plot_gamma(evoked: mne.Evoked, subjects_dir: PathLike = None, **kwargs):
    """Plots the gamma power over time

    Parameters
    ----------
    evoked : mne.Evoked
        The data to plot
    subjects_dir : PathLike, optional
        The subjects directory, by default LAB_root / 'ECoG_Recon_Full'
    **kwargs
        Additional arguments to pass to plot_on_average
        """
    data = evoked.copy().filter(30, 150).apply_hilbert(envelope=True)._data
    fig = plot_on_average(evoked.info, subjects_dir=subjects_dir, **kwargs)
    mne.viz.set_3d_view(fig, azimuth=0, elevation=70)

    xy, im = mne.viz.snapshot_brain_montage(fig, evoked.info)
    # convert from a dictionary to array to plot
    xy_pts = np.vstack([xy[ch] for ch in evoked.info['ch_names']])

    # get a colormap to color nearby points similar colors
    cmap = matplotlib.cm.get_cmap('viridis')

    # create the figure of the brain with the electrode positions
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.set_title('Gamma power over time', size='large')
    ax.imshow(im)
    ax.set_axis_off()

    # normalize gamma power for plotting
    gamma_power = -100 * data / data.max()
    # add the time course overlaid on the positions
    x_line = np.linspace(-0.025 * im.shape[0], 0.025 * im.shape[0],
                         data.shape[1])
    for i, pos in enumerate(xy_pts):
        x, y = pos
        color = cmap(i / xy_pts.shape[0])
        ax.plot(x_line + x, gamma_power[i] + y, linewidth=0.5, color=color)


def get_grey_matter(subjects: Sequence[str]) -> set[str]:
    """Gets the grey matter channels for a list of subjects

    Parameters
    ----------
    subjects : Sequence[str]
        The subjects to get the grey matter channels for

    Returns
    -------
    set[str]
        The grey matter channels
    """
    grey_matter = set()
    for i, subj in enumerate(subjects):
        info = subject_to_info(get_sub(subj))
        parcel_label = gen_labels(info, get_sub(subj))
        subj_grey_matter = pick_no_wm(info.ch_names, parcel_label)

        # get the indices of channels in info that are not in grey_matter
        grey_matter |= {f"{subj}-{ch}" for ch in info.ch_names
                        if ch in subj_grey_matter}
    return grey_matter


def plot_on_average(sigs: Signal | str | mne.Info | list[Signal | str, ...],
                    subj_dir: PathLike = None, rm_wm: bool = False,
                    picks: list[int | str, ...] = None, surface: str = 'pial',
                    hemi: str = 'split', color: matplotlib.colors = (1, 1, 1),
                    size: float = 0.35, fig: Brain = None,
                    label_every: int = None, background: str = 'white',
                    units: str = 'm', transparency: float = 0.6,
                    average: str = 'fsaverage', show: bool = True) -> Brain:
    """Plots the signal on the average brain

    Takes a signal instance or list of signal instances and plots them on the
    fsaverage brain.

    Parameters
    ----------
    sigs : Union[Signal, list[Signal]]
        The signal(s) to plot. If a list, all electrodes will be the same
        color. If a single signal, each electrode will be a different color.
    subj_dir : PathLike, optional
        The subjects directory, by default LAB_root / 'ECoG_Recon'
    rm_wm : bool, optional
        Whether to remove white matter electrodes, by default True
    picks : list[int | str], optional
        The channels to plot, by default None
    surface : str, optional
        The surface to plot on, by default 'pial'
    hemi : str, optional
        The hemisphere to plot, by default 'split'
    color : matplotlib.colors, optional
        The color to plot, by default (1, 1, 1). If None, each electrode group
        will be a different color.
    size : float, optional
        The size of the markers, by default 0.35
    fig : Brain, optional
        The figure to plot on, by default None
    label_every : int, optional
        How often to label the channels, by default None
    background: str, optional
        Background color
    units: str, optional
        Units of the electrodes
    transparency: float, optional
        Transparency of the brain
    average: str, optional
        The average brain to plot on, by default 'fsaverage'
    show: bool, optional
        Whether to show the figure, by default True

    Returns
    -------
    Brain
        The figure brain object
    """

    subj_dir = get_sub_dir(subj_dir)
    if fig is None:
        fig = Brain(average, subjects_dir=subj_dir, cortex='low_contrast',
                    alpha=transparency, background=background, surf=surface,
                    hemi=hemi, units=units, show=show)

    if isinstance(sigs, (Signal, mne.Info)):
        sigs = [sigs]
        color = None
    if isinstance(sigs, Iterable):
        sigs = {get_sub(v): v for v in sigs}

    default_c = parula.mat_colors.copy()
    for subj, inst in sigs.items():

        if isinstance(inst, mne.Info):
            new = inst.copy()
        elif isinstance(inst, Signal):
            new = inst.info.copy()
        elif isinstance(inst, str):
            new = subject_to_info(subj)
            new['subject_info'] = dict(his_id=f"sub-{inst}")
        else:
            raise TypeError(type(inst))

        to_fsaverage = mne.read_talxfm(subj, subj_dir)
        if average == 'fsaverage':
            trans = mne.transforms.Transform(fro='head', to='mri',
                                             trans=to_fsaverage['trans'])
        else:
            from_average = mne.read_talxfm(average, subj_dir)
            to_average = np.dot(np.linalg.inv(from_average['trans']),
                                to_fsaverage['trans'])
            trans = mne.transforms.Transform(fro='head', to='mri',
                                             trans=to_average)

        these_picks = range(len(new.ch_names))
        if isinstance(picks, Iterable):
            if len(picks) == 0:
                continue
            elif isinstance(picks[0], int):
                these_picks = [new.ch_names[pick] for pick in these_picks if
                               pick in picks]
                picks = [p - len(new.ch_names) for p in
                         picks[len(these_picks):]]
            elif isinstance(picks[0], str):
                if '-' in picks[0]:
                    these_picks = [s.split('-')[1] for s in picks if
                                   s.split('-')[0] in new[
                                       'subject_info']['his_id']]
                else:
                    these_picks = [s for s in picks if s in new.ch_names]
        elif picks is not None:
            raise TypeError(picks)

        if len(these_picks) == 0:
            continue

        if rm_wm:
            these_picks = pick_no_wm(these_picks, gen_labels(
                new, subj, subj_dir, new.ch_names))

        if len(these_picks) == 0:
            continue

        # select colors
        if color is None and len(sigs) > 1:
            this_color = []
            p_int = [new.ch_names.index(p) for p in these_picks]
            groups = _group_channels(mne.pick_info(new, p_int))
            n_groups = len(set(groups.values()))
            while len(this_color) < n_groups:
                this_color += [default_c.pop(0)]
        else:
            this_color = color

        # plot the data
        plot_subj(new, subj_dir, these_picks, False, fig=fig,
                  trans=trans, color=this_color, size=size,
                  labels_every=label_every, hemi=hemi, background=background,
                  show=show)

    return fig


def pick_no_wm(picks: list[str], labels: OrderedDict[str: list[str]]) -> list:
    """Picks the channels that are not in the white matter

    Parameters
    ----------
    picks : list[str | int]
        The channels to pick from
    labels : dict[str | int, list[str]]
        The labels for each channel

    Returns
    -------
    list[str | int]
        The channels that are not in the white matter
    """
    bad_words = ('Unknown', 'unknown', 'hypointensities', 'White-Matter')

    # remove corresponding picks with either 'White-Matter' in the left most
    # entry or empty lists
    if isinstance(picks[0], int):
        picks = [list(labels.keys())[p] for p in picks]
    picks = [p for p in picks if not any(w in labels[p] for w in bad_words)]
    return picks


def get_sub(inst: Signal | mne.Info | str) -> str:
    """Gets the subject from the instance

    Parameters
    ----------
    inst : Signal
        The instance to get the subject from

    Returns
    -------
    str
        The subject"""
    if isinstance(inst, Signal):
        inst = inst.info
    elif isinstance(inst, str):
        return f"{inst[0]}{int(inst[1:])}"
    out_str = inst['subject_info']['his_id'][4:]
    if len(out_str) == 1:
        return out_str
    return out_str[0] + str(int("".join(s for s in out_str if s.isdigit())))


# TODO: figure out why elec positions are only correct in meters not mm


def plot_subj(inst: Signal | mne.Info | str, subj_dir: PathLike = None,
              picks: list[str | int] = None, no_wm: bool = False,
              labels_every: int | None = 8, surface: str = 'pial',
              hemi: str = 'both', fig: Brain = None,
              trans=None, color: matplotlib.colors = None,
              size: float = 0.35, show: bool = True, background: str = 'white',
              title: str = None, units: str = 'm', transparency: float = 0.5
              ) -> Brain:
    """Plots the electrodes on the subject's brain

    Parameters
    ----------
    inst : Signal | mne.Info | str
        The subject to plot
    subj_dir : PathLike, optional
        The subjects directory, by default HOME / 'Box' / 'ECoG_Recon'
    picks : list[str | int], optional
        The channels to plot, by default all
    no_wm : bool, optional
        Whether to remove the white matter channels, by default False
    labels_every : int, optional
        How often to label the channels, by default 8
    fig : Brain, optional
        The figure to plot on, by default None
    surface : str, optional
        The surface to plot, by default 'pial'
    hemi : str, optional
        The hemisphere to plot, by default 'split'
    trans: mne.transforms.Transform, optional
        The transformation to apply, by default None
    color : matplotlib.colors, optional
        The color of the electrodes, by default (1,1,1)
    size : float, optional
        The size of the electrodes, by default 0.35
    show : bool, optional
        Whether to show the figure, by default True
    background: str, optional
        Background color
    title : string, optional
        Title the plot
    units: str, optional
        Units of the electrodes
    transparency: float, optional
        Transparency of the brain

    Returns
    -------
    Brain
        The brain plot
    """
    if isinstance(inst, Signal):
        info = inst.info
        sub = get_sub(info)
    elif isinstance(inst, mne.Info):
        info = inst
        sub = get_sub(info)
    elif isinstance(inst, str):
        info = subject_to_info(inst, subj_dir)
        sub = inst
    else:
        raise TypeError(
            f"inst must be Signal, mne.Info, or str, not {type(inst)}")

    if subj_dir is None:
        subj_dir = get_sub_dir(subj_dir)
    if trans is None:
        trans = mne.transforms.Transform(fro='head', to='mri')
    if fig is None:
        fig = Brain(sub, subjects_dir=subj_dir, cortex='low_contrast',
                    alpha=transparency, background=background, surf=surface,
                    hemi=hemi, show=show, units=units)

    if isinstance(color, str):
        color = matplotlib.colors.to_rgb(color)

    # Set the title if provided
    if title is not None:
        mne.viz.set_3d_title(fig, title, size=40)
    if picks is None:
        picks = info.ch_names
    if no_wm:
        picks = pick_no_wm(picks,
                           gen_labels(info, sub, subj_dir, info.ch_names))
    if isinstance(picks[0], str):
        picks = mne.pick_channels(info.ch_names, picks)

    info: mne.Info = mne.pick_info(info, picks)

    # fig.add_sensors(info, trans)
    montage = info.get_montage()
    force2frame(montage, trans.from_str)
    montage.apply_trans(trans)
    pos = {k: v for k, v in montage.get_positions()['ch_pos'].items()}

    # Default montage positions are in m, whereas plotting functions assume mm
    left = {}
    right = {}
    for k, v in pos.items():
        if k.startswith('L'):
            left[k] = v
        elif k.startswith('R'):
            right[k] = v
        elif v[0] < 0:
            left[k] = v
        else:
            right[k] = v

    if left and hemi != 'rh':
        _add_electrodes(fig, info, 'lh', np.vstack(list(left.values())),
                        color, size)
    if right and hemi != 'lh':
        _add_electrodes(fig, info, 'rh', np.vstack(list(right.values())),
                        color, size)

    if labels_every is not None:
        settings = dict(shape=None, always_visible=True,
                        text_color=fig._fg_color, bold=False)
        _add_labels(fig, info, sub, labels_every, hemi,
                    (left, right), **settings)

    return fig


def _add_electrodes(fig: mne.viz.Brain, info: mne.Info, hemi: str,
                    pos: np.ndarray, colors: matplotlib.colors = None,
                    size: float = 0.35):
    groups = _group_channels(info)
    n_groups = len(set(groups.values()))
    if colors is None:
        colors = parula.mat_colors[:n_groups]
    elif (not np.isscalar(colors) and len(colors) != n_groups) \
            or isinstance(colors, tuple):
        colors = [colors] * n_groups
    else:
        colors = list(colors)

    i = 0
    vals = list(groups.values())
    while i < n_groups:
        start, end = vals.index(i), len(vals) - vals[::-1].index(i)
        shank_pos = pos[start:end]
        fig.add_foci(shank_pos, hemi=hemi, color=colors[i],
                     scale_factor=size)
        i += 1


def _group_channels(info, groups: dict = None) -> dict:
    """Automatically find a group based on the name of the channel."""
    if groups is not None:
        for name in info.ch_names:
            if name not in groups.keys():
                raise ValueError(f"{name} not found in ``groups``")
        return groups

    i = 0
    groups = dict()
    base_names = dict()
    for name in info.ch_names:
        # strip all numbers from the name
        base_name = "".join(
            [
                letter
                for letter in name
                if not letter.isdigit() and letter != " "
            ]
        )
        if base_name in base_names:
            # look up group number by base name
            groups[name] = base_names[base_name]
        else:
            groups[name] = i
            base_names[base_name] = i
            i += 1
    return groups


def _add_labels(fig, info, sub, every, hemi, lr, **kwargs):
    names = info.ch_names[slice(every - 1, None, every)]

    if not hemi == 'both':
        for hems, pos in enumerate(lr):
            if (not pos) or \
                    (hemi == 'lh' and hems == 1) or \
                    (hemi == 'rh' and hems == 0):
                continue

            plt_names = filter(lambda x: x.startswith(['L', 'R'][hems]), names)
            plt_names = [f'{sub}-{n}' for n in plt_names]
            positions = np.array([pos[n.split("-")[1]] for n in plt_names])
            fig.plotter.subplot(0, hems)
            fig.plotter.add_point_labels(positions, plt_names, **kwargs)
    else:
        pos = {}
        for hem in lr:
            if hem:
                pos.update(hem)
        plt_names = [f'{sub}-{n}' for n in names]
        positions = np.array([pos[name] for name in names])
        fig.plotter.add_point_labels(positions, plt_names, **kwargs)


def subject_to_info(subject: str, subjects_dir: PathLike = None,
                    ch_types: str = "seeg", sfreq: int = 2000) -> mne.Info:
    """Gets the info for a subject from the subjects directory

    Parameters
    ----------
    subject : str
        The subject to get the info for
    subjects_dir : PathLike, optional
        The subjects directory, by default HOME / 'Box' / 'ECoG_Recon'
    ch_types : str, optional
        The channel type, by default "seeg"
    sfreq : int, optional
        The sampling frequency, by default 2000

    Returns
    -------
    mne.Info
        The info for the subject
    """
    subjects_dir = get_sub_dir(subjects_dir)
    elec_file = op.join(subjects_dir, subject, 'elec_recon',
                        subject + '_elec_locations_RAS_brainshifted.txt')
    elecs = dict()
    with open(elec_file, 'r') as fd:
        reader = csv.reader(fd)
        for row in reader:
            line = row[0].split(" ")
            elecs["".join(line[0:2])] = tuple(
                float(n) / 1000 for n in line[2:5])
    info = mne.create_info(list(elecs.keys()), sfreq, ch_types)
    montage = mne.channels.make_dig_montage(elecs, nasion=(0, 0, 0),
                                            coord_frame='ras')
    info.set_montage(montage)
    return info


@singledispatch
def force2frame(montage: mne.channels.DigMontage, frame: str):
    """Forces the montage to be in the specified frame

    Parameters
    ----------
    montage : mne.channels.DigMontage
        The montage to force
    frame : str, optional
        The frame to force to, by default 'mri'
    """

    settings = dict(fro=montage.get_positions()['coord_frame'],
                    to=frame, trans=np.eye(4))
    # current subjects are in 'mri' space, even though it says head
    if not settings['fro'] == frame:
        trans = mne.transforms.Transform(**settings)
        montage.apply_trans(trans)


@force2frame.register
def _(info: mne.Info, frame: str):
    montage = info.get_montage()
    force2frame(montage, frame)
    info.set_montage(montage)


def gen_labels(info: mne.Info, sub: str = None, subj_dir: PathLike = None,
               picks: list[str] = None) -> OrderedDict[str, list[str]]:
    """Generates the labels for the electrodes

    Parameters
    ----------
    info : mne.Info
        The subject to get the labels for
    subj_dir : PathLike, optional
        The subjects directory, by default None
    picks : list[str | int], optional
        The channels to plot, by default None

    Returns
    -------
    dict[str, list]
        The labels for the electrodes
    """

    sub = get_sub(info) if sub is None else sub
    subj_dir = get_sub_dir(subj_dir)
    montage = info.get_montage()
    force2frame(montage, 'mri')
    # aseg = 'aparc.a2009s+aseg'  # parcellation/anatomical segmentation atlas
    labels = get_elec_volume_labels(sub, subj_dir, 10)

    new_labels = OrderedDict()
    if picks is None:
        picks = info.ch_names

    bad_words = ('Unknown', 'unknown', 'hypointensities', 'White-Matter')
    for p in picks:
        i = 2
        label = labels.T[p].T
        if label[0] not in bad_words:
            new_labels[p] = label[0]
            continue

        while not ((not any(w in label[i] for w in bad_words)) and
                   label[i + 1] > 0.05):
            if (i + 2) <= len(label.T):  # end of labels
                i = 0
                break
            elif label[i + 2].isspace():  # empty label
                i = 0
                break
            else:
                i += 2
        new_labels[p] = label[i]
    return new_labels


if __name__ == "__main__":
    from ieeg.io import get_data, raw_from_layout
    from os import path

    HOME = path.expanduser("~")
    LAB_root = path.join(HOME, "Box", "CoganLab")
    # %% Set up logging
    log_filename = "output.log"
    # op.join(LAB_root, "Aaron_test", "Information.log")
    mne.set_log_file(log_filename,
                     "%(levelname)s: %(message)s - %(asctime)s",
                     overwrite=True)
    mne.set_log_level("INFO")
    TASK = "SentenceRep"
    sub_num = 22
    layout = get_data(TASK, root=LAB_root)
    subj_dir = op.join(LAB_root, "ECoG_Recon_Full")
    sub_pad = "D" + str(sub_num).zfill(4)
    # sub = "D{}".format(sub_num)

    # filt = raw_from_layout(layout.derivatives['clean'], subject=sub_pad,
    #                        extension='.edf', desc='clean', preload=False)

    ##
    # sample_path = mne.datasets.sample.data_path()
    # subjects_dir = sample_path / "subjects"

    brain = plot_subj("D5")
    # fig = plot_on_average(["D24", "D81"], rm_wm=False, hemi='both',
    #                       transparency=0.4,
    #                       picks=list(range(28)) + list(range(52, 176)),
    #                       color=None,
    #                       average="D79", background=(0, 0.4, 0.5))
    # plot_gamma(raw)
    # plot_on_average(["D22", "D28", "D64"],
    #                 picks=["D22-LPIF4", "D28-LPIO7", "D64-LAI6"],
    #                 label_every=1, hemi='lh', rm_wm=False, color='red',
    #                 size=1)



import numpy as np


# %% R-squared (R2)

def get_R2(y_test, y_test_pred):
    """
    Function to get R2

    Parameters
    ----------
    y_test - the true outputs (a matrix of size number of examples x number of
    outputs)
    y_test_pred - the predicted outputs (a matrix of size number of examples x
    number of outputs)

    Returns
    -------
    R2_array: An array of R2s for each output
    """

    # Initialize a list that will contain the R2s for all the outputs
    R2_list = []
    for i in range(y_test.shape[1]):  # Loop through outputs
        # Compute R2 for each output
        y_mean = np.mean(y_test[:, i])
        R2 = 1 - np.sum((y_test_pred[:, i] - y_test[:, i]) ** 2) / np.sum(
            (y_test[:, i] - y_mean) ** 2)
        R2_list.append(R2)  # Append R2 of this output to the list
    R2_array = np.array(R2_list)
    return R2_array  # Return an array of R2s


# %% Pearson's correlation (rho)

def get_rho(y_test, y_test_pred):
    """
    Function to get Pearson's correlation (rho)

    Parameters
    ----------
    y_test - the true outputs (a matrix of size number of examples x number of
    outputs)
    y_test_pred - the predicted outputs (a matrix of size number of examples x
    number of outputs)

    Returns
    -------
    rho_array: An array of rho's for each output
    """
    # Initialize a list that will contain the rhos for all the outputs
    rho_list = []
    for i in range(y_test.shape[1]):  # Loop through outputs
        # Compute rho for each output
        y_mean = np.mean(y_test[:, i])
        rho = np.corrcoef(y_test[:, i].T, y_test_pred[:, i].T)[0, 1]
        rho_list.append(rho)  # Append rho of this output to the list
    rho_array = np.array(rho_list)
    return rho_array  # Return the array of rhos



from os import PathLike as PL

import mne.io
from mne.epochs import BaseEpochs
from mne.evoked import Evoked
from mne.io import base
from numpy import ndarray

PathLike = str | PL
RunDict = dict[int, mne.io.Raw]
SubDict = dict[str, RunDict]
Doubles = tuple[float, float] | list[float, float] | ndarray[(2,), float]
Signal = base.BaseRaw | BaseEpochs | Evoked
ListNum = int | float | ndarray | list | tuple



import gc
from functools import partial

import numpy as np
from joblib import cpu_count
from matplotlib import gridspec
from mne.io import Raw

from ieeg import Doubles, Signal
from ieeg.calc import stats
from ieeg.viz import _qt_backend

_qt_backend()

import matplotlib.pyplot as plt  # noqa: E402


def figure_compare(raw: list[Raw], labels: list[str], avg: bool = True,
                   n_jobs: int = None, **kwargs):
    """Plots the psd of a list of raw objects

    Parameters
    ----------
    raw : list[Raw]
        The raw objects to plot
    labels : list[str]
        The labels for the raw objects
    avg : bool, optional
        Whether to average the psd over channels, by default True
    n_jobs : int, optional
        The number of jobs to use for the computation, by default None
    **kwargs
        Additional keyword arguments to pass to Raw.compute_psd()
    """
    if n_jobs is None:
        n_jobs = cpu_count() - 2
    for title, data in zip(labels, raw):
        title: str
        data: Raw
        psd = data.compute_psd(n_jobs=n_jobs, **kwargs,
                               n_fft=int(data.info['sfreq']))
        fig = psd.plot(average=avg, spatial_colors=avg)
        fig.subplots_adjust(top=0.85)
        fig.suptitle('{}filtered'.format(title), size='xx-large',
                     weight='bold')
        _add_arrows(fig.axes[:2])
        fig.show()
        gc.collect()


def _add_arrows(axes: plt.Axes):
    """add some arrows at 60 Hz and its harmonics"""
    for ax in axes:
        freqs = ax.lines[-1].get_xdata()
        psds = ax.lines[-1].get_ydata()
        for freq in (60, 120, 180, 240):
            idx = np.searchsorted(freqs, freq)
            # get ymax of a small region around the freq. of interest
            y = psds[(idx - 4):(idx + 5)].max()
            ax.arrow(x=freqs[idx], y=y + 18, dx=0, dy=-12, color='red',
                     width=0.1, head_width=3, length_includes_head=True)


def _onclick_select(event, inst, axs):
    for a in axs:
        if event.inaxes == a:
            ch = a.get_title()
            if ch in inst.info['bads']:
                inst.info['bads'].remove(ch)
                print(f"Removing {ch} from bads")
            else:
                inst.info['bads'].append(ch)
                print(f"adding {ch} to bads")


def chan_grid(inst: Signal, n_cols: int = 10, n_rows: int = 6,
              plot_func: callable = None, picks: list[str | int] = None,
              size: tuple[int, int] = (12, 18), show: bool = True, **kwargs
              ) -> list[plt.Figure]:
    """Plot a grid of the channels of a Signal object

    Parameters
    ----------
    size
    inst : Signal
        The Signal object to plot
    n_cols : int, optional
        Number of columns in the grid, by default 10
    n_rows : int, optional
        Number of rows in the grid, by default the minimum number of rows
    plot_func : callable, optional
        The function to use to plot the channels, by default inst.plot()
    picks : list[Union[str, int]], optional
        The channels to plot, by default all
    size : tuple[int, int], optional
        The size of the figure, by default (8, 12)
    show : bool, optional
        Whether to show the figure, by default True

    Returns
    -------
    list[plt.Figure]
        The figures containing the grid
    """
    if n_rows is None:
        n_rows = int(np.ceil(len(inst.ch_names) / n_cols))
    if plot_func is None:
        plot_func = inst.plot
    if picks is None:
        chans = inst.ch_names
    elif isinstance(picks[0], str):
        chans = picks
    elif isinstance(picks[0], int):
        chans = [inst.ch_names[i] for i in picks]
    else:
        raise TypeError("picks must be a list of str or int")

    per_fig = n_cols * n_rows
    numfigs = int(np.ceil(len(chans) / per_fig))
    figs = []
    for i in range(numfigs):
        fig, axs = plt.subplots(nrows=n_rows, ncols=n_cols, frameon=False,
                                figsize=size)

        select = partial(_onclick_select, inst=inst, axs=fig.axes)
        text_spec = dict(fontsize=12, weight="extra bold")

        for j, chan in enumerate(chans[i * per_fig:(i + 1) * per_fig]):
            if j + 1 % n_cols == 0 or i == len(chans) - 1:
                bar = True
            else:
                bar = False
            if "colorbar" in plot_func.__code__.co_varnames:
                kwargs["colorbar"] = bar
            ax = axs.flatten()[j]
            plot_func(picks=[chan], axes=ax, show=False, **kwargs)
            ax.set_title(chan, pad=0, **text_spec)
            ax.tick_params(axis='both', which='major', labelsize=7,
                           direction="in")
            ax.set_xlabel("")
            ax.set_ylabel("")
            gc.collect()
        fig.supxlabel("Time (s)", **text_spec)
        fig.supylabel("Frequency (Hz)", **text_spec)
        if i == numfigs - 1:
            while j + 1 < n_cols * n_rows:
                j += 1
                ax = axs.flatten()[j]
                ax.axis("off")
        fig.canvas.mpl_connect("button_press_event", select)
        fig.tight_layout()
        figs.append(fig)
        if show:
            figs[i].show()
    return figs


def plot_dist(mat: iter, axis: int = 0, mode: str = 'sem',
              mask: np.ndarray = None, times: Doubles = None,
              label: str | int | float = None, color: str | list[int] = None,
              ax: plt.Axes = None) -> plt.Axes:
    """Plot the distribution for a single signal

    A distribution is the mean of the signal over the last dimension, with
    optional masking

    Parameters
    ----------
    mat : iter
        The signal to plot
    axis : int, optional
        The axis to use for the distribution, by default 0
    mode : str, optional
        The mode to use for the distribution, by default 'sem'
    mask : np.ndarray, optional
        The mask to use for the distribution, by default None
    times : Doubles, optional
        The times to use for the x-axis, by default None
    label : Union[str, int, float], optional
        The label for the signal, by default None
    color : Union[str, list[int]], optional
        The color to use for the signal, by default None
    ax : plt.Axes, optional
        The axes to plot on, by default None

    Returns
    -------
    plt.Axes
        The axes containing the plot
        """
    mean, std = stats.dist(mat, axis=axis, where=mask, mode=mode)
    if times is None:
        tscale = range(len(mean))
    else:
        tscale = np.linspace(times[0], times[1], len(mean))
    if ax is None:
        plt.figure()
        ax = plt.gca()
    p = ax.plot(tscale, mean, label=label, color=color)
    if color is None:
        color = p[-1].get_color()
    ax.fill_between(tscale, mean - std, mean + std, alpha=0.2, color=color)
    return ax


def plot_weight_dist(data: np.ndarray, label: np.ndarray, mode: str = 'sem',
                     mask: np.ndarray = None, times: Doubles = None,
                     sig_titles: list[str] = None,
                     colors: list[str | list[int]] = None, ax=None
                     ) -> (plt.Figure, plt.Axes):
    """Basic distribution plot for weighted signals

    Parameters
    ----------
    data : np.ndarray
        The data to plot
    label : np.ndarray
        The labels for the data
    mode : str, optional
        The mode to use for the distribution, by default 'sem'
    mask : np.ndarray, optional
        The mask to use for the distribution, by default None
    times : Doubles, optional
        The times to use for the x-axis
    sig_titles : list[str], optional
        The titles for the signals, by default None
    colors : list[str | list[int]], optional
        The colors for the signals, by default None
    ax : plt.Axes, optional
        The axes to plot on, by default None

    Returns
    -------
    plt.Figure
        The figure containing the plot
    plt.Axes
        The axes containing the plot
    """
    if ax is None:
        fig, ax = plt.subplots()
    if len(label.shape) > 1:
        group = range(min(np.shape(label)))
        weighted = True
    else:
        group = np.unique(label)
        weighted = False
    if sig_titles is None:
        sig_titles = [sig_titles] * len(group)
    if colors is None:
        colors = [colors] * len(group)
    for i, stitle, color in zip(group, sig_titles, colors):
        if not weighted:
            w_sigs = data[label == i]
        else:
            w_sigs = np.multiply(data.T, label[:, i]).T
            # w_sigs = np.average(data, weights=label[:, i], axis=0)
        ax = plot_dist(w_sigs, 0, mode, mask, times, label=stitle,
                       ax=ax, color=color)

    return ax


def subgrids(rows: int, cols: int, sub_cols: int,
             major_rows: tuple[int, ...] = (), titles: list = "",
             ylabels: list = "", xlabels: list = "", **kwargs
             ) -> (plt.Figure, plt.Axes):
    """Create a figure with subgrids

    Parameters
    ----------
    rows : int
        The number of rows in the figure
    cols : int
        The number of columns in the figure
    sub_cols : int
        The number of columns in the subgrids
    major_rows : tuple[int, ...], optional
        The rows that have major subgrids, by default ()
    titles : list, optional
        The titles for the subgrids, by default ""
    ylabels : list, optional
        The ylabels for the subgrids, by default ""
    xlabels : list, optional
        The xlabels for the subgrids, by default ""
    kwargs : dict
        Additional keyword arguments to pass to gridspec

    Returns
    -------
    (plt.Figure, plt.Axes)
        The figure and axes containing the subgrids
    """
    fig = plt.figure()
    gs = fig.add_gridspec(rows, cols, **kwargs)

    labels = dict(title=titles, ylabel=ylabels, xlabel=xlabels)
    for ltype, llist in labels.items():
        if isinstance(llist, str):
            if ltype in ("title", "xlabel"):
                labels[ltype] = [llist] * cols
            else:
                labels[ltype] = [llist] * rows
        elif len(llist) != cols and ltype in ("title", "xlabel"):
            raise ValueError(f"Length of {ltype} must be equal to cols")
        elif len(llist) != rows and ltype in ("ylabel",):
            raise ValueError(f"Length of {ltype} must be equal to rows")

    # Create subplots
    axs = [[None] * cols for _ in range(rows)]
    for r in range(rows):  # Only for the first two rows
        if r in major_rows:
            tc = 1
        else:
            tc = sub_cols
        for c in range(cols):
            gs0 = gs[r, c].subgridspec(1, tc, wspace=0, hspace=0)
            axs[r][c] = gs0.subplots(sharey=True, subplot_kw=dict(frameon=True)
                                     )

            # axes labels
            if r == 0:
                gs0.figure.suptitle(labels["title"][c])
            elif r == rows - 1:
                gs0.figure.supxlabel(labels["xlabel"][c])
            if c == 0:
                gs0.figure.supylabel(labels["ylabel"][r])

    return gs.figure, axs


if __name__ == "__main__":

    gs = subgrids(3, 3, 3, major_rows=(0, 1))



import os

import mne
import vtk

misc_path = mne.datasets.misc.data_path()
sample_path = mne.datasets.sample.data_path()
subjects_dir = sample_path / "subjects"
os.environ["MNE_3D_OPTION_ANTIALIAS"] = "false"

if not os.path.exists(os.path.join(subjects_dir, "1")):
    os.symlink(os.path.join(subjects_dir, 'sample'),
               os.path.join(subjects_dir, '1'))
raw = mne.io.read_raw(misc_path / "seeg" / "sample_seeg_ieeg.fif")


def test_plot_seeg():
    from ieeg.viz.mri import plot_subj
    plot_subj(raw, subj_dir=subjects_dir, show=False)


def test_plot_avg():
    from ieeg.viz.mri import plot_on_average
    plot_on_average(raw, subj_dir=subjects_dir, show=False, rm_wm=False,
                    hemi='both')


# if on windows, finalize closing the vtk window
if os.name == 'nt':
    vtk.vtkRenderWindow().Finalize()
    vtk.vtkObject.GlobalWarningDisplayOff()



import mne
import numpy as np
import pytest
import scipy

from ieeg.calc.fast import mean_diff

bids_root = mne.datasets.epilepsy_ecog.data_path()
seeg = mne.io.read_raw(mne.datasets.misc.data_path() /
                       'seeg' / 'sample_seeg_ieeg.fif')


@pytest.mark.parametrize("mat, axis, expected", [
    # Test case 1: No mask, axis=0
    (np.array([[1, 2, 3], [4, 5, 6]]), 0, (
            [2.5, 3.5, 4.5], [1.5, 1.5, 1.5])),

    # Test case 2: No mask, axis=1
    (np.array([[1, 2, 3], [4, 5, 6]]), 1, (
            [2.0, 5.0], [0.57735027, 0.57735027])),
])
def test_dist(mat, axis, expected):
    from ieeg.calc.stats import dist  # Import your actual module here

    mean, std = dist(mat, axis)
    assert np.allclose(mean, np.array(expected[0]))  # Check mean
    assert np.allclose(std, np.array(expected[1]))  # Check standard deviation


@pytest.mark.parametrize("func, expected", [
    (mean_diff, np.arange(38, 56)),
    (scipy.stats.f_oneway, np.arange(46, 50)),
    (scipy.stats.ttest_ind, np.arange(38, 56))
])
def test_permclust(func, expected):
    from ieeg.navigate import trial_ieeg
    from ieeg.calc import stats

    out = []
    for epoch, t in zip(('Fixation', 'Response'), ((-0.3, 0), (-0.1, 0.2))):
        times = [None, None]
        times[0] = t[0] - 0.5
        times[1] = t[1] + 0.5
        trials = trial_ieeg(seeg, epoch, times, preload=True)
        out.append(trials)
    resp = out[1]
    resp.decimate(10)
    base = out[0]
    base.decimate(10)

    mask, pvals = stats.time_perm_cluster(resp._data[:, 78],
                                          base._data[:, 78], 0.01,
                                          stat_func=func, n_perm=4000)
    assert np.mean(mask[expected]) > 0.8


def test_permclust2():
    from ieeg.calc.stats import time_perm_cluster
    base = np.load("tests/d100_basefield.npy")
    field = base[1]
    base = base[0]
    mask, pvals = time_perm_cluster(field, base, 0.1,
                                    0.1, 10000, 1, 1)
    assert np.any(mask)
    assert np.min(pvals[1]) < 0.05


def test_stats_wavelet():
    from ieeg.navigate import trial_ieeg, outliers_to_nan
    from ieeg.timefreq.utils import wavelet_scaleogram, crop_pad
    from ieeg.calc import stats

    out = []
    for epoch, t in zip(('Fixation', 'Response'), ((-0.3, 0), (-0.1, 0.2))):
        times = [None, None]
        times[0] = t[0] - 0.5
        times[1] = t[1] + 0.5
        trials = trial_ieeg(seeg, epoch, times, preload=True,
                            picks=['LAMY 6', 'LAMY 7'])
        outliers_to_nan(trials, 10)
        spec = wavelet_scaleogram(trials, n_jobs=-2, decim=20)
        crop_pad(spec, "0.5s")
        out.append(spec)
    resp = out[1]
    base = out[0]

    mask, pvals = stats.time_perm_cluster(resp.data, base.data, 0.1,
                                          ignore_adjacency=1, n_perm=10000)
    mask1, pvals1 = stats.time_perm_cluster(resp.data[:, 0], base.data[:, 0],
                                            0.1, n_perm=10000)

    assert np.any(mask)
    assert np.isclose(np.mean(mask1), np.mean(mask[0]))


def test_window_averaged_shuffle():
    from ieeg.calc.stats import window_averaged_shuffle
    base = np.load("tests/d100_basefield.npy")
    field = base[1]
    base = base[0]
    p = window_averaged_shuffle(field, base, 10000, 1, 1)
    t = np.isclose(np.array([0.51315, 0.99920, 6.9993e-04]), p, 0, 0.05)
    assert np.all(t)



import re
from functools import singledispatch
from os import listdir, mkdir, path as op, walk

import mne
import numpy as np
import pandas as pd
from bids import BIDSLayout
from bids.layout import BIDSFile, parse_file_entities
from mne.utils import fill_doc, verbose
from mne_bids import BIDSPath, get_bids_path_from_fname, mark_channels, \
    read_raw_bids, write_raw_bids
from mne_bids.write import _from_tsv

from ieeg import PathLike, Signal


def find_dat(folder: PathLike) -> (PathLike, PathLike):
    """Looks for the .dat file in a specified folder

    Parameters
    ----------
    folder : PathLike
        The folder to search in.

    Returns
    -------
    Tuple[PathLike, PathLike]
        The paths to the ieeg and cleanieeg files.
    """
    cleanieeg = None
    ieeg = None
    for root, _, files in walk(folder):
        for file in files:
            if re.match(r".*cleanieeg\.dat.*", file):
                cleanieeg: PathLike = op.join(root, file)
            elif re.match(r".*ieeg\.dat.*", file):
                ieeg: PathLike = op.join(root, file)
            if ieeg is not None and cleanieeg is not None:
                return ieeg, cleanieeg
    raise FileNotFoundError("Not all .dat files were found:")


def bidspath_from_layout(layout: BIDSLayout, **kwargs) -> BIDSPath:
    """Searches a :class:`BIDSLayout` for a file and returns a
    :class:`BIDSPath` to it.

    Parameters
    ----------
    layout : BIDSLayout
        The :class:`BIDSLayout` to search.
    **kwargs : dict
        The parameters passed to :func:`BIDSLayout.get`

    Returns
    -------
    BIDS_path: BIDSPath
        The BIDSPath to the file.
    """
    my_search: list[BIDSFile] = layout.get(**kwargs)
    if len(my_search) >= 2:
        raise FileNotFoundError("Search terms matched more than one file: \n"
                                "{} \n try adding more search terms"
                                "".format(my_search))
    elif len(my_search) == 0:
        raise FileNotFoundError("No files match your search terms")
    found = my_search[0]
    entities = found.get_entities()
    if 'desc' in entities:
        entities['description'] = entities.pop('desc')
    BIDS_path = BIDSPath(root=layout.root, **entities)
    return BIDS_path


@fill_doc
def raw_from_layout(layout: BIDSLayout, preload: bool = True,
                    run: list[int] | int = None, **kwargs) -> mne.io.Raw:
    """Searches a BIDSLayout for a raw file and returns a mne Raw object.

    Parameters
    ----------
    layout : BIDSLayout
        The BIDSLayout to search.
    %(preload)s
    run : Union[List[int], int], optional
        The run to search for, by default None
    **kwargs : dict
        The parameters passed to bids.BIDSLayout.get()

    Returns
    -------
    mne.io.Raw

    Examples
    --------

    >>> import mne
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=False)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    """
    verbose = kwargs.pop('verbose', True)
    if run is None:
        runs = layout.get(return_type="id", target="run", **kwargs)
    else:
        runs = list(run)
    raw: list[mne.io.Raw] = []
    if runs:
        for r in runs:
            BIDS_path = bidspath_from_layout(layout, run=r, **kwargs)
            new_raw = read_raw_bids(bids_path=BIDS_path, verbose=verbose)
            raw.append(new_raw.copy())
        whole_raw: mne.io.Raw = mne.concatenate_raws(raw)
    else:
        BIDS_path = bidspath_from_layout(layout, **kwargs)
        whole_raw = read_raw_bids(bids_path=BIDS_path, verbose=verbose)
    if preload:
        whole_raw.load_data()
    return whole_raw


def open_dat_file(file_path: str, channels: list[str], sfreq: int = 2048,
                  types: str = "seeg", units: str = "uV") -> mne.io.RawArray:
    """Opens a .dat file and returns a mne.io.RawArray object.

    Parameters
    ----------
    file_path : str
        The path to the .dat file.
    channels : List[str]
        The channels to load.
    sfreq : int, optional
        The sampling frequency, by default 2048
    types : str, optional
        The channel types, by default "seeg"
    units : str, optional
        The units of the data, by default "uV"

    Returns
    -------
    mne.io.RawArray
    """
    with open(file_path, mode='rb') as f:
        data = np.fromfile(f, dtype="float32")
    if "Trigger" in channels:
        channels.remove("Trigger")
    array = np.reshape(data, [len(channels), -1], order='F')
    match units:
        case "V":
            factor = 1
        case "mV":
            factor = 1e-3
        case "uV":
            factor = 1e-6
        case "nV":
            factor = 1e-9
        case _:
            raise NotImplementedError("Unit " + units + " not implemented yet")
    info = mne.create_info(channels, sfreq, types)
    raw = mne.io.RawArray(array * factor, info)
    return raw


def get_data(task: str, root: PathLike) -> BIDSLayout:
    """Gets the data for a subject and task.

    Parameters
    ----------
    task : str, optional
        The task to get the data for, by default "SentenceRep"
    root : PathLike, optional
        The path to the lab directory, by default LAB_root

    Returns
    -------
    layout : BIDSLayout
        The BIDSLayout for the subject.
    """
    BIDS_root = None
    for dir in listdir(root):
        if re.match(r"BIDS-\d\.\d_" + task, dir) and "BIDS" in listdir(op.join(
                root, dir)):
            BIDS_root = op.join(root, dir, "BIDS")
            break
    if BIDS_root is None:
        raise FileNotFoundError("Could not find BIDS directory in {} for task "
                                "{}".format(root, task))
    layout = BIDSLayout(BIDS_root, derivatives=True)
    return layout


@fill_doc
@verbose
def save_derivative(inst: Signal, layout: BIDSLayout, pipeline: str = None,
                    overwrite: bool = False, format: str = 'EDF',
                    anonymize: bool = True, verbose=None):
    """Save an intermediate data instance from a pipeline to a BIDS folder.

    Parameters
    ----------
    inst : Signal
        The data instance to save.
    layout : BIDSLayout
        The BIDSLayout of the original data.
    pipeline : str
        The name of the pipeline.
    %(overwrite)s
    format : str
        The format to save the data in. Defaults to EDF.
    %(verbose)s
    """
    save_dir = op.join(layout.root, "derivatives", pipeline)
    if not op.isdir(save_dir):
        mkdir(save_dir)
    bounds = inst.annotations.copy()
    bounds = bounds[np.where(bounds.description == 'BAD boundary')[0]]
    bounds = [0] + list(bounds.onset) + [inst.times[-1]]
    for i, file in enumerate(inst.filenames):
        entities = parse_file_entities(file)
        if 'desc' in entities.keys():
            entities['description'] = entities.pop('desc')
        if 'subject' not in entities.keys():
            sub = inst.info['subject_info']['his_id']
            entities['subject'] = sub[slice(4, None)]
        if 'task' not in entities.keys():
            entities['task'] = ''
        if pipeline:
            entities['description'] = pipeline
        bids_path = BIDSPath(**entities, root=save_dir)
        run = inst.copy().crop(tmin=bounds[i], tmax=bounds[i + 1])
        if anonymize:
            if isinstance(run, Signal):
                run.anonymize()
                anonymize = None
            else:
                anonymize = {'daysback': 10000000}
        else:
            anonymize = None

        write_raw_bids(run, bids_path, allow_preload=True, format=format,
                       acpc_aligned=True, overwrite=overwrite,
                       anonymize=anonymize, verbose=verbose)


def get_bad_chans(fname: str):
    """Gets the bad channels corresponding to a file.

    Parameters
    ----------
    fname : str
        The path to the file.

    Returns
    -------
    bads : list[str]
        The bad channels.
    """
    data = _from_tsv(fname.replace("_ieeg.edf", "_channels.tsv"))
    bads = [n for n, s in zip(data['name'], data['status']) if s == 'bad']
    return bads


@singledispatch
@verbose
def update(filename: PathLike, channels: list[str],
           description: list[str] | str = None, status: str = 'good',
           verbose=None):
    """Updates the files of a data instance with current metadata

    Parameters
    ----------
    filename : PathLike
        The path to the file to update.
    channels : list[str]
        The channels to update.
    description : list[str] | str, optional
        The description of the channels, by default None
    status : str, optional
        The status of the channels, by default 'good'
    """
    if isinstance(description, str):
        description = [description for _ in range(len(channels))]

    bids_path = get_bids_path_from_fname(filename)
    mark_channels(bids_path, ch_names=channels, status=status,
                  descriptions=description, verbose=verbose)


@update.register
def _(inst: mne.io.base.BaseRaw,
      layout: BIDSLayout, description: list[str] | str = None, verbose=None):
    if not hasattr(inst, 'filenames'):
        inst.filenames = inst.info['subject_info'].get('files', None)
    for i, file in enumerate(inst.filenames):
        fname = op.join(layout.root, file)
        update(fname, inst.info['bads'], description=description, status='bad',
               verbose=verbose)
        goods = [ch for ch in inst.ch_names if ch not in inst.info['bads']]
        update(fname, channels=goods, status='good', verbose=None)


@update.register
def _(inst: mne.time_frequency.BaseTFR,
      layout: BIDSLayout, description: list[str] | str = None, verbose=None):
    if not hasattr(inst, 'filenames'):
        inst.filenames = inst.info['subject_info'].get('files', None)
    for i, file in enumerate(inst.filenames):
        fname = op.join(layout.root, file)
        update(fname, inst.info['bads'], description=description, status='bad',
               verbose=verbose)
        goods = [ch for ch in inst.ch_names if ch not in inst.info['bads']]
        update(fname, channels=goods, status='good', verbose=None)


def get_elec_volume_labels(subj: str, subj_dir: PathLike, radius: int = 3
                           ) -> pd.DataFrame:
    filename = op.join(subj_dir, subj, "elec_recon",
                       f"{subj}_elec_location_radius_{radius}mm_aparc.a2009s+"
                       f"aseg.mgz")
    if op.exists(filename + "_brainshifted.csv"):
        filename += "_brainshifted.csv"
    else:
        filename += ".csv"
    out = pd.read_csv(filename, header=None, skiprows=1, index_col=1)
    return out



import numpy as np


# %% BIN_SPIKES
def bin_spikes(spike_times, dt, wdw_start, wdw_end):
    """
    Function that puts spikes into bins

    Parameters
    ----------
    spike_times: an array of arrays
        an array of neurons. within each neuron's array is an array containing
        all the spike times of that neuron
    dt: number (any format)
        size of time bins
    wdw_start: number (any format)
        the start time for putting spikes in bins
    wdw_end: number (any format)
        the end time for putting spikes in bins

    Returns
    -------
    neural_data: a matrix of size "number of time bins" x "number of neurons"
        the number of spikes in each time bin for each neuron
    """
    edges = np.arange(wdw_start, wdw_end, dt)  # Get edges of time bins
    num_bins = edges.shape[0] - 1  # Number of bins
    num_neurons = spike_times.shape[0]  # Number of neurons
    neural_data = np.empty(
        [num_bins, num_neurons])  # Initialize array for binned neural data
    # Count number of spikes in each bin for each neuron, and put in array
    for i in range(num_neurons):
        neural_data[:, i] = np.histogram(spike_times[i], edges)[0]
    return neural_data


# %% BIN_OUTPUT
def bin_output(outputs, output_times, dt, wdw_start, wdw_end,
               downsample_factor=1):
    """
    Function that puts outputs into bins

    Parameters
    ----------
    outputs: matrix of size "number of times the output was recorded" x "number
    of features in the output"
        each entry in the matrix is the value of the output feature
    output_times: a vector of size "number of times the output was recorded"
        each entry has the time the output was recorded
    dt: number (any format)
        size of time bins
    wdw_start: number (any format)
        the start time for binning the outputs
    wdw_end: number (any format)
        the end time for binning the outputs
    downsample_factor: integer, optional, default=1
        how much to downsample the outputs prior to binning
        larger values will increase speed, but decrease precision

    Returns
    -------
    outputs_binned: matrix of size "number of time bins" x "number of features
    in the output"
        the average value of each output feature in every time bin
    """

    # %% Downsample output
    # We just take 1 out of every "downsample_factor" values#
    if downsample_factor != 1:  # Don't downsample if downsample_factor=1
        # Get the idxs of values we are going to include after downsampling
        downsample_idxs = np.arange(0, output_times.shape[0],
                                    downsample_factor)
        outputs = outputs[downsample_idxs, :]  # Get the downsampled outputs
        output_times = output_times[
            downsample_idxs]  # Get the downsampled output times

    # %% Put outputs into bins
    edges = np.arange(wdw_start, wdw_end, dt)  # Get edges of time bins
    num_bins = edges.shape[0] - 1  # Number of bins
    output_dim = outputs.shape[1]  # Number of output features
    outputs_binned = np.empty(
        [num_bins, output_dim])  # Initialize matrix of binned outputs
    # Loop through bins, and get the mean outputs in those bins
    for i in range(num_bins):  # Loop through bins
        # Indices to consider the output signal (in the correct time range)
        idxs = np.where((np.squeeze(output_times) >= edges[i]) & (
                np.squeeze(output_times) < edges[i + 1]))[0]
        for j in range(output_dim):  # Loop through output features
            outputs_binned[i, j] = np.mean(outputs[idxs, j])

    return outputs_binned


# %% GET_SPIKES_WITH_HISTORY
def get_spikes_with_history(neural_data, bins_before, bins_after,
                            bins_current=1):
    """
    Function that creates the covariate matrix of neural activity

    Parameters
    ----------
    neural_data: a matrix of size "number of time bins" x "number of neurons"
        the number of spikes in each time bin for each neuron
    bins_before: integer
        How many bins of neural data prior to the output are used for decoding
    bins_after: integer
        How many bins of neural data after the output are used for decoding
    bins_current: 0 or 1, optional, default=1
        Whether to use the concurrent time bin of neural data for decoding

    Returns
    -------
    X: a matrix of size "number of total time bins" x "number of surrounding
    time bins used for prediction" x "number of neurons"
        For every time bin, there are the firing rates of all neurons from the
        specified number of time bins before (and after)
    """

    num_examples = neural_data.shape[
        0]  # Number of total time bins we have neural data for
    num_neurons = neural_data.shape[1]  # Number of neurons
    surrounding_bins = bins_before + bins_after + bins_current
    # Number of surrounding time bins used for prediction
    X = np.empty([num_examples, surrounding_bins,
                  num_neurons])  # Initialize covariate matrix with NaNs
    X[:] = np.NaN
    # Loop through each time bin, and collect the spikes occurring in
    # surrounding time bins. Note that the first "bins_before" and last
    # "bins_after" rows of X will remain filled with NaNs, since they don't get
    # filled in below. This is because, for example, we cannot collect 10 time
    # bins of spikes before time bin 8
    start_idx = 0
    # The first bins_before and last bins_after bins don't get filled in
    for i in range(num_examples - bins_before - bins_after):
        # The bins of neural data we will be including are between start_idx
        # and end_idx (which will have length "surrounding_bins")
        end_idx = start_idx + surrounding_bins
        # Put neural data from surrounding bins in X, starting at row
        # "bins_before"
        X[i + bins_before, :, :] = neural_data[start_idx:end_idx, :]
        start_idx = start_idx + 1
    return X



import numpy as np


def stitch_mats(mats: list[np.ndarray], overlaps: list[int], axis: int = 0
                ) -> np.ndarray:
    """break up the matrices into their overlapping and non-overlapping parts
    then stitch them back together

    Parameters
    ----------
    mats : list
        The matrices to stitch together
    overlaps : list
        The number of overlapping rows between each matrix
    axis : int, optional
        The axis to stitch along, by default 0

    Returns
    -------
    np.ndarray
        The stitched matrix

    Examples
    --------
    >>> mat1 = np.array([[1, 2, 3], [4, 5, 6]])
    >>> mat2 = np.array([[7, 8, 9], [10, 11, 12]])
    >>> mat3 = np.array([[13, 14, 15], [16, 17, 18]])
    >>> stitch_mats([mat1, mat2, mat3], [1, 1])
    array([[ 1,  2,  3],
           [10, 11, 12],
           [16, 17, 18]])
    >>> stitch_mats([mat1, mat2, mat3], [0, 0], axis=1)
    array([[ 1,  2,  3,  7,  8,  9, 13, 14, 15],
           [ 4,  5,  6, 10, 11, 12, 16, 17, 18]])
    >>> mat4 = np.array([[19, 20, 21], [22, 23, float("nan")]])
    >>> stitch_mats([mat3, mat4], [0], axis=1)
    array([[13., 14., 15., 19., 20., 21.],
           [16., 17., 18., 22., 23., nan]])
    """
    stitches = [mats[0]]
    if len(mats) != len(overlaps) + 1:
        raise ValueError("The number of matrices must be one more than the num"
                         "ber of overlaps")
    for i, over in enumerate(overlaps):
        stitches = stitches[:-2] + merge(stitches[-1], mats[i + 1], over, axis)
    out = np.concatenate(stitches, axis=axis)
    if np.array_equal(out.astype(int), out):
        return out.astype(int)
    else:
        return out


def merge(mat1: np.ndarray, mat2: np.ndarray, overlap: int, axis: int = 0
          ) -> list[np.ndarray[float]]:
    """Take two arrays and merge them over the overlap gradually"""
    sl = [slice(None)] * mat1.ndim
    sl[axis] = slice(0, mat1.shape[axis] - overlap)
    start = mat1[tuple(sl)]
    sl[axis] = slice(mat1.shape[axis] - overlap, mat1.shape[axis])
    middle1 = np.multiply(np.linspace(1, 0, overlap), mat1[tuple(sl)])
    sl[axis] = slice(0, overlap)
    middle2 = np.multiply(np.linspace(0, 1, overlap), mat2[tuple(sl)])
    middle = np.add(middle1, middle2)
    sl[axis] = slice(overlap, mat2.shape[axis])
    last = mat2[tuple(sl)]

    return [start, middle, last]


def make_data_same(data_fix: np.ndarray, shape: tuple | list,
                   stack_ax: int = 0, pad_ax: int = -1) -> np.ndarray:
    """Force the last dimension of data_fix to match the last dimension of
    shape.

    Takes the two arrays and checks if the last dimension of data_fix is
    smaller than the last dimension of shape. If there's more than two
    dimensions, it will rearrange the data to match the shape. If there's only
    two dimensions, it will repeat the signal to match the shape of data_like.
    If the last dimension of data_fix is larger than the last dimension of
    shape, it will return a subset of data_fix.

    Parameters
    ----------
    data_fix : array
        The data to reshape.
    shape : list | tuple
        The shape of data to match.

    Returns
    -------
    data_fix : array
        The reshaped data.

    Examples
    --------
    >>> np.random.seed(0)
    >>> data_fix = np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]])
    >>> make_data_same(data_fix, (2, 8))
    array([[ 1,  2,  3,  4,  5,  4,  3,  2],
           [ 6,  7,  8,  9, 10,  9,  8,  7]])
    >>> (newarr := make_data_same(data_fix, (2, 2)))
    array([[1, 2],
           [3, 4],
           [6, 7],
           [8, 9]])
    >>> make_data_same(newarr, (3, 2), stack_ax=1, pad_ax=0)
    array([[1, 2],
           [3, 4],
           [6, 7]])
    """

    stack_ax, pad_ax = list(range(len(shape)))[stack_ax], \
        list(range(len(shape)))[pad_ax]

    # Check if the pad dimension of data_fix is smaller than the pad
    # dimension of shape
    if data_fix.shape[pad_ax] <= shape[pad_ax]:
        return pad_to_match(np.zeros(shape), data_fix, stack_ax)

    # When the pad dimension of data_fix is larger than the pad dimension of
    # shape, take subsets of data_fix and stack them together on the stack
    # dimension
    else:
        return rand_offset_reshape(data_fix, shape, stack_ax, pad_ax)


def pad_to_match(sig1: np.ndarray, sig2: np.ndarray,
                 axis: int | tuple[int, ...] = 0) -> np.ndarray:
    """Pad the second signal to match the first signal along all axes not
    specified.

    Takes the two arrays and checks if the shape of sig2 is smaller than the
    shape of sig1. For each axis not specified, it will pad the second signal
    to match the first signal along that axis.

    Parameters
    ----------
    sig1 : array
        The data to match.
    sig2 : array
        The data to pad.
    axis : int | tuple
        The axes along which to pad the data.

    Returns
    -------
    sig2 : array
        The padded data.
    """
    # Make sure the data is the same shape
    if np.isscalar(axis):
        axis = (axis,)
    axis = list(axis)
    for i, ax in enumerate(axis):
        axis[i] = np.arange(sig1.ndim)[ax]
    eq = list(e for i, e in enumerate(np.equal(sig1.shape, sig2.shape))
              if i not in axis)
    if not all(eq):
        for ax in axis:
            eq.insert(ax, True)
        pad_shape = [(0, 0) if eq[i] else
                     (0, sig1.shape[i] - sig2.shape[i])
                     for i in range(sig1.ndim)]
        sig2 = np.pad(sig2, pad_shape, mode='reflect')
    return sig2


def rand_offset_reshape(data_fix: np.ndarray, shape: tuple, stack_ax: int,
                        pad_ax: int) -> np.ndarray:
    """Take subsets of data_fix and stack them together on the stack dimension

    This function takes the data and reshapes it to match the shape by taking
    subsets of data_fix and stacking them together on the stack dimension,
    randomly offsetting the start of the first subset. It is assumed that the
    padding axis 'pad_ax' is larger in data_fix.shape than in shape.

    Parameters
    ----------
    data_fix : array
        The data to reshape.
    shape : list | tuple
        The shape of data to match.
    stack_ax : int
        The axis along which to stack the subsets.
    pad_ax : int
        The axis along which to slice the subsets.

    Returns
    -------
    data_fix : array
        The reshaped data.

    Examples
    --------
    >>> np.random.seed(0)
    >>> data_fix = np.arange(50).reshape((5, 10))
    >>> data_fix
    array([[ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24, 25, 26, 27, 28, 29],
           [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
           [40, 41, 42, 43, 44, 45, 46, 47, 48, 49]])
    >>> rand_offset_reshape(data_fix, (2, 4), 0, 1)
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [10, 11, 12, 13],
           [14, 15, 16, 17],
           [20, 21, 22, 23],
           [24, 25, 26, 27],
           [30, 31, 32, 33],
           [34, 35, 36, 37],
           [40, 41, 42, 43],
           [44, 45, 46, 47]])
    >>> rand_offset_reshape(data_fix, (2, 4), 1, 0) # doctest: +ELLIPSIS
    array([[ 0, 20,  1, 21,  2, 22,  3, 23,  4, 24,  5, 25,  6, 26,  7, 27,
             8, 28,  9, 29],
           [10, 30, 11, 31, 12, 32, 13, 33, 14, 34, 15, 35, 16, 36, 17, 37,
            18, 38, 19, 39]])
    """

    # Randomly offset the start of the first subset
    num_stack = data_fix.shape[pad_ax] // shape[pad_ax]
    if data_fix.shape[pad_ax] % shape[pad_ax] == 0:
        num_stack -= 1
    offset = np.random.randint(0, data_fix.shape[pad_ax] - shape[
        pad_ax] * num_stack)

    # Create an array to store the output
    out_shape = [shape[i] if i == pad_ax else data_fix.shape[i]
                 for i in range(data_fix.ndim)]
    out_shape[stack_ax] *= num_stack
    out = np.zeros(tuple(out_shape), dtype=data_fix.dtype)

    # Iterate over the subsets
    sl_in = [slice(None)] * data_fix.ndim
    sl_out = [slice(None)] * data_fix.ndim
    for i in range(num_stack):
        # Get the start and end indices of the subset
        start = i * shape[pad_ax] + offset
        end = start + shape[pad_ax]

        # Create a slice object for the subset
        sl_in[pad_ax] = slice(start, end)
        sl_out[stack_ax] = slice(i, None, num_stack)

        # Fill in the subset
        out[tuple(sl_out)] = data_fix[tuple(sl_in)]

    return out



from typing import Literal, Tuple

import numpy as np
from numpy.typing import NDArray
from sklearn.model_selection import RepeatedStratifiedKFold

import itertools
from ieeg.calc.fast import mixup, norm

Array2D = NDArray[Tuple[Literal[2], ...]]
Vector = NDArray[Literal[1]]


class MinimumNaNSplit(RepeatedStratifiedKFold):
    """A Repeated Stratified KFold iterator that splits the data into sections

    This class splits the data into sections, checking that the training set
    never has fewer than the specified number of non-NaN values.
    Parameters
    ----------
    n_splits : int
        The number of splits.
    n_repeats : int, optional
        The number of times to repeat the splits, by default 10.
    random_state : int, optional
        The random state to use, by default None.

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(0)
    >>> X = np.vstack((np.arange(1, 9).reshape(4, 2), np.full((4, 2), np.nan)))
    >>> y = np.array([0, 0, 1, 1, 0, 0, 1, 1])
    >>> msn = MinimumNaNSplit(2, 3)
    >>> for train, test in msn.split(X, y):
    ...     print("train:", train, "test:", test)
    train: [1 3 4 6] test: [0 2 5 7]
    train: [0 2 5 7] test: [1 3 4 6]
    train: [1 3 4 7] test: [0 2 5 6]
    train: [0 2 5 6] test: [1 3 4 7]
    train: [0 3 5 7] test: [1 2 4 6]
    train: [1 2 4 6] test: [0 3 5 7]
    >>> msn = MinimumNaNSplit(2, 3, which='test', min_non_nan=1)
    >>> for train, test in msn.split(X, y):
    ...     print("train:", train, "test:", test)
    train: [2 3 4 5] test: [0 1 6 7]
    train: [0 1 6 7] test: [2 3 4 5]
    train: [0 1 2 7] test: [3 4 5 6]
    train: [3 4 5 6] test: [0 1 2 7]
    train: [3 4 5 7] test: [0 1 2 6]
    train: [0 1 2 6] test: [3 4 5 7]
    """

    def __init__(self, n_splits: int, n_repeats: int = 10,
                 random_state: int = None, min_non_nan: int = 2,
                 which: str = 'train'):
        super().__init__(n_splits=n_splits, n_repeats=n_repeats,
                         random_state=random_state)
        self.n_splits = n_splits
        self.min_non_nan = min_non_nan
        if which not in ('train', 'test'):
            raise ValueError("which must be either 'train' or 'test'")
        self.which = which

    def split(self, X, y=None, groups=None):

        # find where the nans are
        where = np.isnan(X).any(axis=tuple(range(X.ndim))[1:])
        not_where = np.where(~where)[0]
        where = np.where(where)[0]

        # if there are no nans, then just split the data
        if len(where) == 0:
            yield from super(MinimumNaNSplit, self).split(X, y, groups)
            return
        elif (n_non_nan := not_where.shape[0]) < (n_min := self.min_non_nan +
                                                  1):
            raise ValueError(f"Need at least {n_min} non-nan values, but only"
                             f" have {n_non_nan}")

        check = {'train': lambda t: np.setdiff1d(not_where, t,
                                                 assume_unique=True),
                 'test': lambda t: np.intersect1d(not_where, t,
                                                  assume_unique=True)}

        splits = super().split(X, y, groups)

        # check that all training sets for each kfold within each repetition
        # have at least min_non_nan non-nan values
        kfold_set = [None] * self.n_splits
        while element := next(splits, False):
            for i in range(self.n_splits):
                if i == 0:
                    train, test = element
                else:
                    train, test = next(splits)

                # if any test set has more non-nan values than the total number
                # of non-nan values minus the minimum number of non-nan values,
                # then throw out the split and append an extra repetition
                if check[self.which](test).shape[0] < self.min_non_nan:
                    for _ in range(i + 1, self.n_splits):
                        next(splits)
                    extra = super().split(X, y, groups)
                    one_rep = itertools.islice(extra, self.n_splits)
                    splits = itertools.chain(one_rep, splits)
                    break
                kfold_set[i] = (train, test)
            else:
                yield from kfold_set

    @staticmethod
    def oversample(arr: np.ndarray, func: callable = mixup,
                   axis: int = 1, copy: bool = True, seed=None) -> np.ndarray:
        """Oversample nan rows using func

        Parameters
        ----------
        arr : array
            The data to oversample.
        func : callable
            The function to use to oversample the data.
        axis : int
            The axis along which to apply func.
        copy : bool
            Whether to copy the data before oversampling.

        Examples
        --------
        >>> np.random.seed(0)
        >>> arr = np.array([[1, 2], [4, 5], [7, 8],
        ... [float("nan"), float("nan")]])
        >>> MinimumNaNSplit.oversample(arr, norm, 0)
        array([[1.        , 2.        ],
               [4.        , 5.        ],
               [7.        , 8.        ],
               [8.32102813, 5.98018098]])
        >>> MinimumNaNSplit.oversample(arr, mixup, 0, seed=42) # doctest: +SKIP
        array([[1.        , 2.        ],
               [4.        , 5.        ],
               [7.        , 8.        ],
               [5.24946679, 6.24946679]])
        """
        if copy:
            arr = arr.copy()

        axis = arr.ndim + axis if axis < 0 else axis

        if arr.ndim <= 0:
            raise ValueError("Cannot apply func to a 0-dimensional array")
        elif seed is not None:
            func(arr, axis, seed=seed)
        else:
            func(arr, axis)
        return arr

    def shuffle_labels(self, arr: np.ndarray, labels: np.ndarray,
                       trials_ax: int = 0, min_trials: int = 1):
        """Shuffle the labels while making sure that the minimum non nan
        trials are kept

        Parameters
        ----------
        arr : array
            The data to shuffle.
        labels : array
            The labels to shuffle.
        trials_ax : int
            The axis along which to apply func.
        min_trials : int
            The minimum number of non-nan trials to keep. By default,
            self.n_splits

        Examples
        --------
        >>> np.random.seed(0)
        >>> arr = np.array([[[1, 2], [4, 5], [7, 8],
        ... [float("nan"), float("nan")]]])
        >>> labels = np.array([0, 0, 1, 1])
        >>> MinimumNaNSplit(1).shuffle_labels(arr, labels, 1, 1)
        >>> labels
        array([1, 1, 0, 0])
        """
        cats = np.unique(labels)
        gt_labels = [0] * cats.shape[0]
        min_trials *= self.n_splits
        i = 0
        while not all(g >= min_trials for g in gt_labels):
            np.random.shuffle(labels)
            for j, l in enumerate(cats):
                eval_arr = np.take(arr, np.flatnonzero(labels == l), trials_ax)
                gt_labels[j] = np.min(np.sum(
                    np.all(~np.isnan(eval_arr), axis=2), axis=trials_ax))
            if sum(gt_labels) < min_trials * cats.shape[0]:
                raise ValueError("Not enough non-nan trials to shuffle")
            i += 1
            if i > 100000:
                raise ValueError("Could not find a shuffle that satisfies the"
                                 " minimum number of non-nan trials "
                                 f"{gt_labels}")


def oversample_nan(arr: np.ndarray, func: callable, axis: int = 1,
                   copy: bool = True, seed: int = None) -> np.ndarray:
    """Oversample nan rows using func

    Parameters
    ----------
    arr : array
        The data to oversample.
    func : callable
        The function to use to oversample the data.
    axis : int
        The axis along which to apply func.
    copy : bool
        Whether to copy the data before oversampling.

    Examples
    --------
    >>> np.random.seed(0)
    >>> arr = np.array([[1, 2], [4, 5], [7, 8],
    ... [float("nan"), float("nan")]])
    >>> oversample_nan(arr, norm, 0)
    array([[1.        , 2.        ],
           [4.        , 5.        ],
           [7.        , 8.        ],
           [8.32102813, 5.98018098]])
    >>> oversample_nan(arr, mixup, 0, seed=42) # doctest: +SKIP
    array([[1.        , 2.        ],
           [4.        , 5.        ],
           [7.        , 8.        ],
           [5.24946679, 6.24946679]])
    >>> arr3 = np.arange(24, dtype=float).reshape(2, 3, 4)
    >>> arr3[0, 2, :] = [float("nan")] * 4
    >>> oversample_nan(arr3, mixup, 1, seed=42) # doctest: +SKIP
    array([[[ 0.        ,  1.        ,  2.        ,  3.        ],
            [ 4.        ,  5.        ,  6.        ,  7.        ],
            [ 2.33404428,  3.33404428,  4.33404428,  5.33404428]],
    <BLANKLINE>
           [[12.        , 13.        , 14.        , 15.        ],
            [16.        , 17.        , 18.        , 19.        ],
            [20.        , 21.        , 22.        , 23.        ]]])
    >>> oversample_nan(arr3, norm, 1)
    array([[[ 0.        ,  1.        ,  2.        ,  3.        ],
            [ 4.        ,  5.        ,  6.        ,  7.        ],
            [ 3.95747597,  7.4817864 ,  7.73511598,  3.04544424]],
    <BLANKLINE>
           [[12.        , 13.        , 14.        , 15.        ],
            [16.        , 17.        , 18.        , 19.        ],
            [20.        , 21.        , 22.        , 23.        ]]])
    """

    if copy:
        arr = arr.copy()

    axis = arr.ndim + axis if axis < 0 else axis

    if arr.ndim <= 0:
        raise ValueError("Cannot apply func to a 0-dimensional array")
    elif func is mixup and seed is not None:
        func(arr, axis, seed=seed)
    else:
        func(arr, axis)
    return arr


def find_nan_indices(arr: np.ndarray, obs_axis: int) -> tuple:
    """Find the indices of rows with and without NaN values

    Parameters
    ----------

    arr : array
        The data to find indices.
    obs_axis : int
        The axis along which to apply func.

    Returns
    -------
    tuple
        A tuple of two arrays containing the indices of rows with and without
        NaN values.

    Examples
    --------
    >>> arr = np.array([[1, 2], [4, 5], [7, 8],
    ... [float("nan"), float("nan")]])
    >>> find_nan_indices(arr, 0) # doctest: +ELLIPSIS
    (array([3]... array([0, 1, 2]...)
    """

    # Initialize boolean mask of rows with NaN values
    not_obs = tuple(i for i in range(arr.ndim) if i != obs_axis)

    # Check each row individually
    nan = np.any(np.isnan(arr), axis=not_obs)

    # Get indices of rows with and without NaN values using boolean indexing
    nan_rows = np.flatnonzero(nan)
    non_nan_rows = np.flatnonzero(~nan)

    return nan_rows, non_nan_rows


def sortbased_rand(n_range: int, iterations: int, n_picks: int = -1):
    """Generate random numbers using sort-based sampling

    Parameters
    ----------
    n_range : int
        The range of numbers to sample from.
    iterations : int
        The number of iterations to run.
    n_picks : int
        The number of numbers to pick from the range. If -1, then the number of
        picks is equal to the range.

    Returns
    -------
    array
        An array of shape (iterations, n_picks) containing the random numbers.

    References
    ----------
    [1] `stackoverflow link <https://stackoverflow.com/questions/31955660/effic
    iently-generating-multiple-instances-of-numpy-random-choice-without-replace
    /31958263#31958263>`_
    """
    return np.argsort(np.random.rand(iterations, n_range), axis=1
                      )[:, :n_picks]



""" Cross-Patient Decoder Classes

Author: Zac Spalding
"""

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.decomposition import PCA


class crossPtDecoder(BaseEstimator):

    def preprocess_train(self, X, y=None):
        pass

    def preprocess_test(self, X, y=None):
        pass

    def fit(self, X, y, **kwargs):
        X_p, y_p = self.preprocess_train(X, y, **kwargs)
        return self.decoder.fit(X_p, y_p)

    def predict(self, X):
        X_p = self.preprocess_test(X)
        return self.decoder.predict(X_p)

    def score(self, X, y, **kwargs):
        X_p = self.preprocess_test(X)
        return self.decoder.score(X_p, y, **kwargs)


class crossPtDecoder_sepDimRed(crossPtDecoder):
    """ Cross-Patient Decoder with separate PCA for each patient. """

    def __init__(self, cross_pt_data, decoder, dim_red=PCA, n_comp=10):
        self.cross_pt_data = cross_pt_data
        self.decoder = decoder
        self.dim_red = dim_red
        self.n_comp = n_comp

    def preprocess_train(self, X, y, **kwargs):
        cross_pt_trials = [x.shape[0] for x, _, _ in self.cross_pt_data]
        # reshape features to be 2D (preserve last dimension for reduction)
        X_cross_r = [x.reshape(-1, x.shape[-1]) for x, _, _ in
                     self.cross_pt_data]
        X_tar_r = X.reshape(-1, X.shape[-1])
        # reduce dimensionality of cross-patient data
        X_cross_dr = [self.dim_red(n_components=self.n_comp).fit_transform(x)
                      for x in X_cross_r]

        # reduce dimensionality of target data, saving dim. red. object for
        # test set
        tar_dr = self.dim_red(n_components=self.n_comp)
        X_tar_dr = tar_dr.fit_transform(X_tar_r)
        self.tar_dr = tar_dr

        # reshape for concatenation
        X_cross_dr = [x.reshape(cross_pt_trials[i], -1, x.shape[-1]) for i, x
                      in enumerate(X_cross_dr)]
        X_cross_dr = [x.reshape(x.shape[0], -1) for x in X_cross_dr]
        X_tar_dr = X_tar_dr.reshape(X.shape[0], -1)

        # concatenate cross-patient data
        X_dr = np.vstack([X_tar_dr] + X_cross_dr)
        y_dr = np.hstack([y] + [y for _, y, _ in self.cross_pt_data])
        return X_dr, y_dr

    def preprocess_test(self, X):
        X_r = X.reshape(-1, X.shape[-1])
        X_dr = self.tar_dr.transform(X_r)
        return X_dr.reshape(X.shape[0], -1)


class crossPtDecoder_sepAlign(crossPtDecoder):

    def __init__(self, cross_pt_data, decoder, aligner, dim_red=PCA,
                 n_comp=10):
        self.cross_pt_data = cross_pt_data
        self.decoder = decoder
        self.dim_red = dim_red
        self.n_comp = n_comp
        self.aligner = aligner

    def preprocess_train(self, X, y, y_align=None):
        cross_pt_trials = [x.shape[0] for x, _, _ in self.cross_pt_data]
        tar_trials = X.shape[0]
        # reshape features to be 2D (preserve last dimension for reduction)
        X_cross_r = [x.reshape(-1, x.shape[-1]) for x, _, _ in
                     self.cross_pt_data]
        X_tar_r = X.reshape(-1, X.shape[-1])
        # reduce dimensionality of cross-patient data
        X_cross_dr = [self.dim_red(n_components=self.n_comp).fit_transform(x)
                      for x in X_cross_r]

        # reduce dimensionality of target data, saving dim. red. object for
        # test set
        tar_dr = self.dim_red(n_components=self.n_comp)
        X_tar_dr = tar_dr.fit_transform(X_tar_r)
        self.tar_dr = tar_dr

        # reshape back to 3D
        X_cross_dr = [x.reshape(cross_pt_trials[i], -1, x.shape[-1]) for i, x
                      in enumerate(X_cross_dr)]
        X_tar_dr = X_tar_dr.reshape(X.shape[0], -1, X_tar_dr.shape[-1])

        # option for separate alignment labels
        if y_align is None:
            y_align = y
        y_align_cross = [y_a for _, _, y_a in self.cross_pt_data]

        algns = [self.aligner() for _ in range(len(self.cross_pt_data))]
        X_algn_dr = []
        for i, algn in enumerate(algns):
            algn.fit(X_tar_dr, X_cross_dr[i], y_align, y_align_cross[i])
            X_algn_dr.append(algn.transform(X_cross_dr[i]))

        X_algn_dr = [x.reshape(x.shape[0], -1) for x in X_algn_dr]
        X_tar_dr = X_tar_dr.reshape(X_tar_dr.shape[0], -1)

        # concatenate cross-patient data
        X_pool = np.vstack([X_tar_dr] + X_algn_dr)
        y_pool = np.hstack([y] + [y for _, y, _ in self.cross_pt_data])
        return X_pool, y_pool

    def preprocess_test(self, X):
        X_r = X.reshape(-1, X.shape[-1])
        X_dr = self.tar_dr.transform(X_r)
        return X_dr.reshape(X.shape[0], -1)



import numpy as np
import pytest

from ieeg.calc.mat import LabeledArray, Labels, combine, iter_nest_dict
from ieeg.calc.fast import concatenate_arrays


@pytest.mark.parametrize("arrays, axis, expected_output", [
    # Test case 1: Concatenate along axis 0
    ((np.array([]), np.array([[1, 2], [3, 4]]),
      np.array([[5, 6, 7], [8, 9, 10]])),
     0,
     np.array([[1, 2, np.nan], [3, 4, np.nan], [5, 6, 7], [8, 9, 10]])),

    # Test case 2: Concatenate along axis 1
    ((np.ones((2, 1)), np.zeros((3, 1))), 1,
     np.array([[1, 0], [1, 0], [np.nan, 0]])),

    # Test case 3: Empty input arrays
    ((np.array([]), np.array([])), 0, None),

    # Test case 4: Concatenate along axis 2
    ((np.array([[[1]], [[2]]]), np.array([[[3], [4]], [[5], [6]]])),
     2,
     np.array([[[1, 3], [np.nan, 4]], [[2, 5], [np.nan, 6]]])),

    # Test case 5: Concatenate along axis 0 with empty array in the middle
    ((np.array([[1, 2], [3, 4]]), np.array([]),
      np.array([[5, 6, 7], [8, 9, 10]])),
     0,
     np.array([[1, 2, np.nan], [3, 4, np.nan], [5, 6, 7], [8, 9, 10]])),

    # Test case 6: Concatenate along axis 0 with empty arrays at the beginning
    # and end
    ((np.array([]), np.array([[1, 2], [3, 4]]), np.array([])),
     0,
     np.array([[1, 2], [3, 4]])),

    # Test case 7: Concatenate along axis -1 (last axis)
    ((np.array([[[1]], [[2]]]), np.array([[[3], [4]], [[5], [6]]])),
     -1,
     np.array([[[1, 3], [np.nan, 4]], [[2, 5], [np.nan, 6]]])),

    # Test case 8: Concatenate an array containing only nan values
    ((np.array([[np.nan, np.nan], [np.nan, np.nan]]),
      np.array([[5, 6, 7], [8, 9, 10]])),
     0,
     np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan],
               [5, 6, 7], [8, 9, 10]])),

    # Test case 9: Concatenate along new axis
    # ((np.array([[1, 2], [3, 4]]), np.array([[5, 6, 7], [8, 9, 10]])),
    #     None,
    #     np.array([[[1, 2, np.nan], [3, 4, np.nan]], [[5, 6, 7],
    #     [8, 9, 10]]]))
])
def test_concatenate_arrays(arrays, axis, expected_output):
    print(f"Shapes {[arr.shape for arr in arrays]}")
    try:
        new = concatenate_arrays(arrays, axis)
        print(f"New shape {new.shape}")
        if axis is None:
            axis = 0
            arrays = tuple(np.expand_dims(arr, axis) for arr in arrays)
        while axis < 0:
            axis += new.ndim
        assert np.array_equal(new, expected_output, True)
    except ValueError as e:
        try:
            assert expected_output is None
        except AssertionError:
            raise e


# Test creation of ArrayDict
def test_array_creation():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    assert isinstance(ad, LabeledArray)


# Test conversion to numpy array
def test_array_to_array():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    np_array = np.array([[[1, 2, 3], [4, 5, np.nan]]])
    assert np.array_equal(ad, np_array, True)


# Test index parsing
@pytest.mark.parametrize('index, expected', [
    (('a', 'b', 'c'), (0, 0, 0)),
    ((slice(None), ('b', 'f')), (range(1), [0, 1], range(3))),
    ((..., 'f', (1, 3)), (range(1), 1, [1, 3])),
    (('g'), "g not found in ['a']"),
    ((slice(None), np.array([False, True])), (range(1), [False, True],
                                              range(3)))
])
def test_parse_index(index, expected):
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3},
                  'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    try:
        parsed = ad._parse_index(list(index))
        parsed = tuple(tuple(p) if isinstance(p, np.ndarray)
                       else p for p in parsed)
    except IndexError as e:
        parsed = e.args[0]
    assert parsed == expected


# Test getting all keys
def test_array_all_keys():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    keys = [Labels(['a']), Labels(['b', 'f']), Labels(['c', 'd', 'e'])]
    assert all(np.array_equal(ad.labels[i], keys[i]) for i in range(len(keys)))


# Test getting all keys in a really nested ArrayDict
def test_array_all_keys_nested():
    data = {'a': {'b': {'c': {'d': {'e': {'f': {'g': {'h': {'i': {'j': {
        'k': 1}}}}}}}}}}}
    ad = LabeledArray.from_dict(data)
    keys = [['a'], ['b'], ['c'], ['d'], ['e'], ['f'], ['g'], ['h'], ['i'],
            ['j'], ['k']]
    assert ad.labels == keys


# Test getting all keys in a large ArrayDict (10000 keys)
def test_array_all_keys_large():
    data = {str(i): i for i in range(100000)}
    ad = LabeledArray.from_dict(data)
    labels = set(ad.labels[0])
    assert labels == set(map(str, range(100000)))


# Test indexing with a single key
def test_array_single_key_indexing():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    subset = LabeledArray.from_dict({'c': 1, 'd': 2, 'e': 3})
    assert np.array_equal(ad['a']['b'], subset)


# Test indexing with a tuple of keys that leads to a scalar value
def test_array_scalar_value_indexing():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    assert ad['a']['b']['d'] == 2


# Test shape property
def test_array_shape():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    assert ad.shape == (1, 2, 3)


# Test combine
@pytest.mark.parametrize('data, dims, expected', [
    ({'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}, (1, 2),
     {'a': {'b-c': 1, 'b-d': 2, 'b-e': 3, 'f-c': 4, 'f-d': 5}}),
    ({'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}, (0, 2),
     {'b': {'a-c': 1, 'a-d': 2, 'a-e': 3}, 'f': {'a-c': 4, 'a-d': 5}}),
    ({'a': {'b': {'c': {'d': {'e': 1, 'f': 2}}}}}, (0, 4),
     {'b': {'c': {'d': {'a-e': 1, 'a-f': 2}}}}),
    ({'a': {'b': {'c': {'d': {'e': 1, 'f': 2}}}},
      'g': {'b': {'c': {'d': {'e': 3, 'f': 4}}}}}, (0, 4),
     {'b': {'c': {'d': {'a-e': 1, 'a-f': 2, 'g-e': 3, 'g-f': 4}}}})
])
def test_combine(data, dims, expected):
    new = combine(data, dims)
    assert new == expected


# Test nested dict iterator
def test_dict_iterator():
    data = {'a': {'b': {'c': {'d': {'e': 1, 'f': 2}}}}}
    iterator = iter_nest_dict(data)
    assert len(list(iterator)) == 2


# Test combine dimensions with arrays
def test_array_dict_combine_dimensions_with_arrays():
    data = {'b': {'c': np.array([1, 2, 3]), 'd': np.array([4, 5, 6])},
            'f': {'c': np.array([7, 8, 9])}}
    ad = LabeledArray.from_dict(data)
    new = ad.combine((1, 2))
    assert new == LabeledArray([[1., 2., 3., 4., 5., 6.],
                                [7., 8., 9., np.nan, np.nan, np.nan]],
                               labels=[('b', 'f'), ('c-0', 'c-1', 'c-2',
                                                    'd-0', 'd-1', 'd-2')])
    assert new['b'].to_dict() == {'c-0': 1., 'c-1': 2., 'c-2': 3., 'd-0': 4.,
                                  'd-1': 5., 'd-2': 6.}


def test_from_dict():
    data = {'a': {'b': {'c': 1, 'd': 2, 'e': 3}, 'f': {'c': 4, 'd': 5}}}
    ad = LabeledArray.from_dict(data)
    expected_labels = [['a'], ['b', 'f'], ['c', 'd', 'e']]
    expected_array = LabeledArray([[[1, 2, 3], [4, 5, float('nan')]]],
                                  expected_labels)
    assert ad == expected_array


def test_eq():
    data1 = {'a': {'b': {'c': 1}}}
    ad1 = LabeledArray.from_dict(data1)

    data2 = {'a': {'b': {'c': 1}}}
    ad2 = LabeledArray.from_dict(data2)

    data3 = {'a': {'b': {'d': 1}}}
    ad3 = LabeledArray.from_dict(data3)

    assert ad1 == ad2
    assert ad1 != ad3


def test_repr():
    data = {'a': {'b': {'c': 1}}}
    ad = LabeledArray.from_dict(data)
    expected_repr = ("array([[[1.]]])\nlabels(['a']\n       ['b']\n       ['c'"
                     "])")
    assert repr(ad) == expected_repr


@pytest.mark.parametrize('idx', [
    (0,),
    (0, 0),
    (..., 0, 0),
    (..., 0),
    (slice(None), 0),
    (slice(None), 0, ...)
])
def test_numpy_idx(idx):
    data = np.array([[1., 2., 3.], [4., 5., 6.]])
    ad = LabeledArray(data, labels=(('a', 'b'), ('c', 'd', 'e')))
    assert np.array_equal(ad[idx], data[idx])


@pytest.mark.parametrize('idx, expected', [
    ((0,), (('b',), ('c', 'd'))),
    ((0, 0), (('c', 'd'),)),
    ((..., 0, 0), (('a',),)),
    ((..., 0), (('a',), ('b',))),
    ((slice(None), 0), (('a',), ('c', 'd'))),
    ((slice(None), 0, slice(None)), (('a',), ('c', 'd'))),
    ((slice(None), 'b'), (('a',), ('c', 'd'))),
    ((slice(None), 'b', 'c'), (('a',),)),
])
def test_idx(idx, expected):
    ad = LabeledArray([[[1, 2]]], labels=[('a',), ('b',), ('c', 'd')])
    expected = [Labels(ex) for ex in expected]
    assert all((ad[idx].labels[i] == expected[i]).all() for i in
               range(len(expected)))


@pytest.mark.parametrize('idx, val, expected', [
    ((0, 0, 0), 3, [[[3, 2]]]),
    ((0, 0, 1), 3, [[[1, 3]]]),
    (('a', 'b', 'c'), 3, [[[3, 2]]]),
    (('a', 'b', 'd'), 3, [[[1, 3]]]),
    ((0, 0), [3, 4], [[[3, 4]]]),
    ((0, 'b'), [3, 4], [[[3, 4]]])
])
def test_set_array_val(idx, val, expected):
    ad = LabeledArray([[[1, 2]]], labels=[('a',), ('b',), ('c', 'd')])
    ad[idx] = val
    np.testing.assert_array_equal(ad.__array__(), np.array(expected))



import mne
import numpy as np
from bids import BIDSLayout
from mne.utils import fill_doc, verbose
from scipy.signal import detrend

from ieeg import Doubles, Signal
from ieeg.calc import stats
from ieeg.io import update
from ieeg.timefreq.utils import to_samples


def crop_empty_data(raw: mne.io.Raw, bound: str = 'boundary',
                    start_pad: str = "10s", end_pad: str = "10s"
                    ) -> mne.io.Raw:
    """Crops out long stretches of data with no events.

    Takes raw instance with annotated events and crops the instance so that the
    raw file starts at start_pad before the first event and stops an amount of
    time in seconds given by end_pad after the last event.

    Parameters
    ----------
    raw : mne.io.Raw
        The raw file to crop.
    bound : str, optional
        The annotation description to use as a boundary, by default 'boundary'
    start_pad : str, optional
        The amount of time to pad the start of the file, by default "10s"
    end_pad : str, optional
        The amount of time to pad the end of the file, by default "10s"

    Returns
    -------
    mne.io.Raw
        The cropped raw file.

    Examples
    --------
    >>> import mne
    >>> from ieeg.io import raw_from_layout
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=False)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> cropped = crop_empty_data(raw, 'onset')
    >>> cropped.times[0], cropped.times[-1]
    (0.0, 104.94)
    """

    crop_list = []

    start_pad = to_samples(start_pad, raw.info['sfreq']) / raw.info['sfreq']
    end_pad = to_samples(end_pad, raw.info['sfreq']) / raw.info['sfreq']

    # split annotations into blocks
    annot = raw.annotations.copy()
    block_idx = [idx + 1 for idx, val in
                 enumerate(annot) if bound in val['description']]
    block_annot = [annot[i: j] for i, j in
                   zip([0] + block_idx, block_idx +
                       ([len(annot)] if block_idx[-1] != len(annot) else []))]

    for block_an in block_annot:
        # remove boundary events from annotations
        no_bound = None
        for an in block_an:
            if bound not in an['description']:
                if no_bound is None:
                    no_bound = mne.Annotations(**an)
                else:
                    an.pop('orig_time')
                    no_bound.append(**an)

        # Skip if block is all boundary events
        if no_bound is None:
            continue
        # get start and stop time from raw.annotations onset attribute
        t_min = max(0, no_bound.onset[0] - start_pad)
        t_max = no_bound.onset[-1] + end_pad

        # create new cropped raw file
        crop_list.append(raw.copy().crop(tmin=t_min, tmax=t_max))

    return mne.concatenate_raws(crop_list)


@fill_doc
@verbose
def channel_outlier_marker(input_raw: Signal, outlier_sd: float = 3,
                           max_rounds: int = np.inf, axis: int = 0,
                           save: bool = False, verbose: bool = True
                           ) -> list[str]:
    """Identify bad channels by variance.

    Parameters
    ----------
    input_raw : Signal
        Raw data to be analyzed.
    outlier_sd : int, optional
        Number of standard deviations above the mean to be considered an
        outlier, by default 3
    max_rounds : int, optional
        Maximum number of variance estimations, by default runs until no
        more bad channels are found.
    axis : int, optional
        Axis to calculate variance over, by default 0
    save : bool, optional
        Whether to save bad channels to raw.info['bads'], by default False
    %(verbose)s

    Returns
    -------
    list[str]
        List of bad channel names.

    Examples
    --------
    >>> import mne
    >>> from ieeg.io import raw_from_layout
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=False)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> bads = channel_outlier_marker(raw, 3, 2)
    outlier round 1 channels: ['AST2']
    outlier round 1 channels: ['AST2', 'RQ2']
    outlier round 1 channels: ['AST2', 'RQ2', 'N/A']
    outlier round 2 channels: ['AST2', 'RQ2', 'N/A', 'G32']
    outlier round 2 channels: ['AST2', 'RQ2', 'N/A', 'G32', 'AD3']
    outlier round 2 channels: ['AST2', 'RQ2', 'N/A', 'G32', 'AD3', 'PD4']
    """

    tmp = input_raw.copy()
    data = detrend(tmp.get_data('data'))  # channels X time
    names = tmp.pick('data').ch_names
    bads = []  # output for bad channel names
    desc = []  # output for bad channel descriptions

    # Pop out names to bads output using comprehension list
    for ind, i in stats.outlier_repeat(data, outlier_sd, max_rounds, axis):
        bads.append(names[ind])
        desc.append(f'outlier round {i} more than {outlier_sd} SDs above mean')
        # log channels excluded per round
        if verbose:
            mne.utils.logger.info(f'outlier round {i} channels: {bads}')

    if save:
        if not hasattr(tmp, 'filenames'):
            raise ValueError("Raw instance must have filenames attribute to "
                             "save bad channels")
        for file in tmp.filenames:
            update(file, bads, desc)

    return bads


@verbose
def outliers_to_nan(trials: mne.epochs.BaseEpochs, outliers: float,
                    copy: bool = False, picks: list = 'data',
                    verbose=None) -> mne.epochs.BaseEpochs:
    """Set outliers to nan.

    Parameters
    ----------
    trials : mne.epochs.BaseEpochs
        The trials to remove outliers from.
    outliers : float
        The number of standard deviations above the mean to be considered an
        outlier.
    copy : bool, optional
        Whether to copy the data, by default False
    picks : list, optional
        The channels to remove outliers from, by default 'data'

    Returns
    -------
    mne.epochs.BaseEpochs
        The trials with outliers set to nan.

    Examples
    --------
    >>> import mne
    >>> from ieeg.io import raw_from_layout
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=50)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> epochs = trial_ieeg(raw, "AD1-4, ATT1,2", (-1, 2), preload=True,
    ... verbose=False)
    >>> epochs = outliers_to_nan(epochs, 3, verbose=False)
    >>> epochs['AD1-4, ATT1,2'].get_data()[0]
    array([[        nan,         nan,         nan, ...,         nan,
                    nan,         nan],
           [-0.00030586, -0.00030625, -0.00031171, ..., -0.00016054,
            -0.00015976, -0.00015664],
           [        nan,         nan,         nan, ...,         nan,
                    nan,         nan],
           ...,
           [-0.00021483, -0.00021131, -0.00023084, ..., -0.00034295,
            -0.00032381, -0.00031444],
           [-0.00052188, -0.00052852, -0.00053125, ..., -0.00046211,
            -0.00047148, -0.00047891],
           [-0.00033708, -0.00028005, -0.00020934, ..., -0.00040934,
            -0.00042341, -0.00040973]])
    """
    if copy:
        trials = trials.copy()
    picks = mne.io.pick._picks_to_idx(trials.info, picks)
    trials.load_data()
    data = trials.get_data(picks=picks, verbose=verbose)

    # bool array of where to keep data trials X channels
    keep = stats.find_outliers(data, outliers)

    # set outliers to nan if not keep
    data = np.where(keep[..., None], data, np.nan)
    trials._data[:, picks] = data

    return trials


@fill_doc
@verbose
def trial_ieeg(raw: mne.io.Raw, event: str | list[str, ...], times: Doubles,
               verbose=None, **kwargs) -> mne.Epochs:
    """Epochs data from a mne Raw iEEG instance.

    Takes a mne Raw instance and epochs the data around a specified event. If
    baseline is specified, the data is also epoched around the baseline event
    and the baseline is subtracted from the data epochs.

    Parameters
    ----------
    raw : mne.io.Raw
        The raw data to epoch.
    event : str
        The event to epoch around.
    times : tuple[float, float]
        The time window to epoch around the event.
    %(picks_all)s
    %(reject_epochs)s
    %(flat)s
    %(decim)s
    %(epochs_reject_tmin_tmax)s
    %(detrend_epochs)s
    %(proj_epochs)s
    %(on_missing_epochs)s
    %(verbose)s

    Returns
    -------
    mne.Epochs
        The epoched data.

    Examples
    --------
    >>> import mne
    >>> from ieeg.io import raw_from_layout
    >>> bids_root = mne.datasets.epilepsy_ecog.data_path(verbose=False)
    >>> layout = BIDSLayout(bids_root)
    >>> raw = raw_from_layout(layout, subject="pt1", preload=True,
    ... extension=".vhdr", verbose=False)
    Reading 0 ... 269079  =      0.000 ...   269.079 secs...
    >>> epochs = trial_ieeg(raw, "AD1-4, ATT1,2", (-1, 2), verbose=True
    ... ) # doctest: +ELLIPSIS
    Used Annotations descriptions: ['AD1-4, ATT1,2', 'AST1,3', 'G16', 'PD',...
    Not setting metadata
    1 matching events found
    No baseline correction applied
    0 projection items activated
    >>> epochs = trial_ieeg(raw, ['AST1,3', 'G16'], (-1, 2), verbose=True
    ... ) # doctest: +ELLIPSIS
    Used Annotations descriptions: ['AD1-4, ATT1,2', 'AST1,3', 'G16', 'PD', ...
    Not setting metadata
    2 matching events found
    No baseline correction applied
    0 projection items activated
    """

    # determine the events
    events, ids = mne.events_from_annotations(raw)
    dat_ids = [ids[i] for i in mne.event.match_event_names(ids, event)]
    rev = {k: v for k, v in ids.items() if v in dat_ids}

    # epoch the data
    return mne.Epochs(raw, events, event_id=rev, tmin=times[0],
                      tmax=times[1], baseline=None, verbose=verbose, **kwargs)


if __name__ == "__main__":
    from os import path
    from ieeg.io import raw_from_layout

    # %% Set up logging
    log_filename = "output.log"
    # op.join(LAB_root, "Aaron_test", "Information.log")
    mne.set_log_file(log_filename,
                     "%(levelname)s: %(message)s - %(asctime)s",
                     overwrite=True)
    mne.set_log_level("INFO")
    HOME = path.expanduser("~")
    LAB_root = path.join(HOME, "Box", "CoganLab")
    TASK = "SentenceRep"
    sub_num = 29
    subj = "D" + str(sub_num).zfill(4)
    # layout, raw, D_dat_raw, D_dat_filt = get_data(sub_num, TASK)
    bids_root = LAB_root + "/BIDS-1.0_SentenceRep/BIDS"
    layout = BIDSLayout(bids_root, derivatives=True)
    filt = raw_from_layout(layout.derivatives['filt'], subject=subj,
                           extension='.edf', desc='filt', preload=True)
    raw = raw_from_layout(layout, subject=subj, extension='.edf', desc=None,
                          preload=True)
    events, event_id = mne.events_from_annotations(filt)
    auds = mne.Epochs(filt, events, event_id['Audio'], baseline=None, tmin=-2,
                      tmax=5, preload=True, detrend=1)
    bads = channel_outlier_marker(auds)
    auds.info['bads'] = bads
